"""
Train several models all at once.

HiveTrainer
we could make a framework that trains multiple LLMs all at once on separate threads and they fight for compute
time. This would be a good way to train a lot of models at once. We could also make it so that the models can
communicate with each other and share information

Features:
    * Train several models all at once
    * Models can communicate with each other and share information
    * Models fight for compute time, so the best model wins
    * Model with highest loss is killed and replaced with a new model
    * Models can be trained on separate GPUs
    * Multiple ModelS can be trained on a single GPU


"""

import threading

from zeta.training.train import Trainer


def model_training_thread(model, *args, **kwargs):
    """Model training thread with model and args and kwargs"""
    trainer = Trainer(model=model, *args, **kwargs)
    trainer.train()


class HiveTrainer:
    """
    HiveTrainer

    Train several models all at once.

    HiveTrainer
    we could make a framework that trains multiple LLMs all at once on separate threads and they fight for compute
    time. This would be a good way to train a lot of models at once. We could also make it so that the models can
    communicate with each other and share information

    Features:
        * Train several models all at once
        * Models can communicate with each other and share information
        * Models fight for compute time, so the best model wins
        * Model with highest loss is killed and replaced with a new model
        * Models can be trained on separate GPUs
        * Multiple ModelS can be trained on a single GPU

    Args:
        models (list): List of models to train
        gradient_accumluate_every (int): Gradient accumulate every
        batch_size (int): Batch size
        seq_len (int): Sequence length
        entity_name (str): Entity name
        use_fsdp (bool): Use FSDP
        use_activation_checkpointing (bool): Use activation checkpointing
        learning_rate (float): Learning rate
        seed (int): Seed
        use_pretokenized (bool): Use pretokenized
        resume_from_checkpoint (str): Resume from checkpoint
        checkpointing_steps (int): Checkpointing steps
        output_dir (str): Output directory
        weight_decay (float): Weight decay
        use_deepspeed (bool): Use deepspeed

    Methods:
        train: Train the models

    Usage:
        # Instantiate models
        models = [YourModelClass1(), YourModelClass2()]  # Replace with your model classes





    """

    def __init__(
        self,
        models,
        gradient_accumluate_every,
        batch_size,
        seq_len,
        entity_name,
        use_fsdp=False,
        use_activation_checkpointing=False,
        learning_rate=None,
        seed=None,
        use_pretokenized=False,
        resume_from_checkpoint=None,
        checkpointing_steps=None,
        output_dir=None,
        weight_decay=None,
        use_deepspeed=None,
    ):
        self.models = models
        self.gradient_accumluate_every = gradient_accumluate_every
        self.batch_size = batch_size
        self.seq_len = seq_len
        self.entity_name = entity_name
        self.use_fsdp = use_fsdp
        self.use_activation_checkpointing = use_activation_checkpointing
        self.learning_rate = learning_rate
        self.seed = seed
        self.use_pretokenized = use_pretokenized
        self.resume_from_checkpoint = resume_from_checkpoint
        self.checkpointing_steps = checkpointing_steps
        self.output_dir = output_dir
        self.weight_decay = weight_decay
        self.use_deepspeed = use_deepspeed

    def train(
        self,
        MASTER_ADDR=None,
        MASTER_PORT=None,
        WORLD_SIZE=None,
        RANK=None,
    ):
        """
        Train the models

        Args:
            MASTER_ADDR (str): Master address
            MASTER_PORT (str): Master port
            WORLD_SIZE (str): World size
            RANK (str): Rank

        Usage:
            hive_trainer.train(MASTER_ADDR='localhost', MASTER_PORT='9994', RANK='0', WORLD_SIZE=str(torch.cuda.device_count()))

        """
        threads = []

        for model in self.models:
            t = threading.Thread(
                target=model_training_thread,
                args=(model,),
                kwargs={
                    "gradient_accumulate_every": self.gradient_accumulate_every,
                    "batch_size": self.batch_size,
                    "seq_len": self.seq_len,
                    "entity_name": self.entity_name,
                    "use_fsdp": self.use_fsdp,
                    "use_activation_checkpointing": (
                        self.use_activation_checkpointing
                    ),
                    "learning_rate": self.learning_rate,
                    "seed": self.seed,
                    "use_pretokenized": self.use_pretokenized,
                    "resume_from_checkpoint": self.resume_from_checkpoint,
                    "checkpointing_steps": self.checkpointing_steps,
                    "output_dir": self.output_dir,
                    "weight_decay": self.weight_decay,
                    "use_deepspeed": self.use_deepspeed,
                    "MASTER_ADDR": MASTER_ADDR,
                    "MASTER_PORT": MASTER_PORT,
                    "RANK": RANK,
                    "WORLD_SIZE": WORLD_SIZE,
                },
            )
            t.start()
            threads.append(t)

        for t in threads:
            t.join()


# # Instantiate models
# models = [YourModelClass1(), YourModelClass2()]  # Replace with your model classes

# # Instantiate HiveTrainer and begin training
# hive_trainer = HiveTrainer(
#     models=models,
#     gradient_accumulate_every=1,
#     batch_size=32,
#     seq_len=512,
#     entity_name="my_project",
#     learning_rate=0.001,
#     seed=42
# )

# hive_trainer.train(MASTER_ADDR='localhost', MASTER_PORT='9994', RANK='0', WORLD_SIZE=str(torch.cuda.device_count()))



import logging
import math
from collections import OrderedDict

import torch
from torch import Tensor, nn

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class PytorchGELUTanh(nn.Module):
    """
    A fast C implementation of the tanh approximation of the GeLU activation function. See
    https://arxiv.org/abs/1606.08415.

    This implementation is equivalent to NewGELU and FastGELU but much faster. However, it is not an exact numerical
    match due to rounding errors.
    """

    def __init__(self):
        super().__init__()

    def forward(self, input: Tensor) -> Tensor:
        return nn.functional.gelu(input, approximate="tanh")


class NewGELUActivation(nn.Module):
    """
    Implementation of the GELU activation function currently in Google BERT repo (identical to OpenAI GPT). Also see
    the Gaussian Error Linear Units paper: https://arxiv.org/abs/1606.08415
    """

    def forward(self, input: Tensor) -> Tensor:
        return (
            0.5
            * input
            * (
                1.0
                + torch.tanh(
                    math.sqrt(2.0 / math.pi)
                    * (input + 0.044715 * torch.pow(input, 3.0))
                )
            )
        )


class GELUActivation(nn.Module):
    """
    Original Implementation of the GELU activation function in Google BERT repo when initially created. For
    information: OpenAI GPT's GELU is slightly different (and gives slightly different results): 0.5 * x * (1 +
    torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3)))) This is now written in C in nn.functional
    Also see the Gaussian Error Linear Units paper: https://arxiv.org/abs/1606.08415
    """

    def __init__(self, use_gelu_python: bool = False):
        super().__init__()
        if use_gelu_python:
            self.act = self._gelu_python
        else:
            self.act = nn.functional.gelu

    def _gelu_python(self, input: Tensor) -> Tensor:
        return input * 0.5 * (1.0 + torch.erf(input / math.sqrt(2.0)))

    def forward(self, input: Tensor) -> Tensor:
        return self.act(input)


class FastGELUActivation(nn.Module):
    """
    Applies GELU approximation that is slower than QuickGELU but more accurate. See: https://github.com/hendrycks/GELUs
    """

    def forward(self, input: Tensor) -> Tensor:
        return (
            0.5
            * input
            * (
                1.0
                + torch.tanh(
                    input * 0.7978845608 * (1.0 + 0.044715 * input * input)
                )
            )
        )


class QuickGELUActivation(nn.Module):
    """
    Applies GELU approximation that is fast but somewhat inaccurate. See: https://github.com/hendrycks/GELUs
    """

    def forward(self, input: Tensor) -> Tensor:
        return input * torch.sigmoid(1.702 * input)


class ClippedGELUActivation(nn.Module):
    """
    Clip the range of possible GeLU outputs between [min, max]. This is especially useful for quantization purpose, as
    it allows mapping negatives values in the GeLU spectrum. For more information on this trick, please refer to
    https://arxiv.org/abs/2004.09602.

    Gaussian Error Linear Unit. Original Implementation of the gelu activation function in Google Bert repo when
    initially created.

    For information: OpenAI GPT's gelu is slightly different (and gives slightly different results): 0.5 * x * (1 +
    torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3)))). See https://arxiv.org/abs/1606.08415
    """

    def __init__(self, min: float, max: float):
        if min > max:
            raise ValueError(
                f"min should be < max (got min: {min}, max: {max})"
            )

        super().__init__()
        self.min = min
        self.max = max

    def forward(self, x: Tensor) -> Tensor:
        return torch.clip(gelu(x), self.min, self.max)


class AccurateGELUActivation(nn.Module):
    """
    Applies GELU approximation that is faster than default and more accurate than QuickGELU. See:
    https://github.com/hendrycks/GELUs

    Implemented along with MEGA (Moving Average Equipped Gated Attention)
    """

    def __init__(self):
        super().__init__()
        self.precomputed_constant = math.sqrt(2 / math.pi)

    def forward(self, input: Tensor) -> Tensor:
        return (
            0.5
            * input
            * (
                1
                + torch.tanh(
                    self.precomputed_constant
                    * (input + 0.044715 * torch.pow(input, 3))
                )
            )
        )


class MishActivation(nn.Module):
    """
    See Mish: A Self-Regularized Non-Monotonic Activation Function (Misra., https://arxiv.org/abs/1908.08681). Also
    visit the official repository for the paper: https://github.com/digantamisra98/Mish
    """

    def __init__(self):
        super().__init__()
        self.act = nn.functional.mish

    def _mish_python(self, input: Tensor) -> Tensor:
        return input * torch.tanh(nn.functional.softplus(input))

    def forward(self, input: Tensor) -> Tensor:
        return self.act(input)


class LinearActivation(nn.Module):
    """
    Applies the linear activation function, i.e. forwarding input directly to output.
    """

    def forward(self, input: Tensor) -> Tensor:
        return input


class LaplaceActivation(nn.Module):
    """
    Applies elementwise activation based on Laplace function, introduced in MEGA as an attention activation. See
    https://arxiv.org/abs/2209.10655

    Inspired by squared relu, but with bounded range and gradient for better stability
    """

    def forward(self, input, mu=0.707107, sigma=0.282095):
        input = (input - mu).div(sigma * math.sqrt(2.0))
        return 0.5 * (1.0 + torch.erf(input))


class ReLUSquaredActivation(nn.Module):
    """
    Applies the relu^2 activation introduced in https://arxiv.org/abs/2109.08668v2
    """

    def forward(self, input):
        relu_applied = nn.functional.relu(input)
        squared = torch.square(relu_applied)
        return squared


class ClassInstantier(OrderedDict):
    def __getitem__(self, key):
        content = super().__getitem__(key)
        cls, kwargs = content if isinstance(content, tuple) else (content, {})
        return cls(**kwargs)


ACT2CLS = {
    "gelu": GELUActivation,
    "gelu_10": (ClippedGELUActivation, {"min": -10, "max": 10}),
    "gelu_fast": FastGELUActivation,
    "gelu_new": NewGELUActivation,
    "gelu_python": (GELUActivation, {"use_gelu_python": True}),
    "gelu_pytorch_tanh": PytorchGELUTanh,
    "gelu_accurate": AccurateGELUActivation,
    "laplace": LaplaceActivation,
    "leaky_relu": nn.LeakyReLU,
    "linear": LinearActivation,
    "mish": MishActivation,
    "quick_gelu": QuickGELUActivation,
    "relu": nn.ReLU,
    "relu2": ReLUSquaredActivation,
    "relu6": nn.ReLU6,
    "sigmoid": nn.Sigmoid,
    "silu": nn.SiLU,
    "swish": nn.SiLU,
    "tanh": nn.Tanh,
}
ACT2FN = ClassInstantier(ACT2CLS)


def get_activation(activation_string):
    if activation_string in ACT2FN:
        return ACT2FN[activation_string]
    else:
        raise KeyError(
            f"function {activation_string} not found in ACT2FN mapping"
            f" {list(ACT2FN.keys())}"
        )


# For backwards compatibility with: from activations import gelu_python
gelu_python = get_activation("gelu_python")
gelu_new = get_activation("gelu_new")
gelu = get_activation("gelu")
gelu_fast = get_activation("gelu_fast")
quick_gelu = get_activation("quick_gelu")
silu = get_activation("silu")
mish = get_activation("mish")
linear_act = get_activation("linear")



import torch.distributed as dist  # Add this line


def print_num_params(model):
    """Print the number of parameters in a model.

    Args:
        model (_type_): _description_
    """
    n_params = sum(p.numel() for p in model.parameters() if p.requires_grad)

    if dist.is_available():
        if dist.get_rank() == 0:
            print(f"Number of parameters in model: {n_params}")
    else:
        print(f"Number of parameters in model: {n_params}")


def print_main(msg):
    """Print the message only on the main process.

    Args:
        msg (_type_): _description_
    """
    if dist.is_available():
        if dist.get_rank() == 0:
            print(msg)
    else:
        print(msg)



import torch
import torch.nn.functional as F
from torch import nn

from zeta.nn.attention import FlashAttention


class ModalityAdaptiveModule(nn.Module):
    """
    Modality Adaptive Module

    Args:
        dim: int
            The dimension of the input features
        heads: int
            The number of heads to use for the attention mechanism

    Returns:
        x: torch.Tensor


    Examples:
        >>> x = torch.randn(1, 3, 512)
        >>> y = torch.randn(1, 3, 512)
        >>> model = ModalityAdaptiveModule(512, 8)
        >>> out = model(x, y)
        >>> print(out.shape)
        torch.Size([1, 3, 512])


    """

    def __init__(self, dim: int, heads: int, dropout: float = 0.1):
        super().__init__()
        self.dim = dim
        self.heads = heads
        self.dropout = dropout
        self.scale = dim**-0.5
        assert dim % heads == 0, "dim must alwasy be divisible by heads"

        # Initialize the normalization layers for each modality
        self.norm_text = nn.LayerNorm(dim)
        self.norm_img = nn.LayerNorm(dim)

        # Initialize the img linear layers
        self.img_v_proj = nn.Linear(dim, dim)
        self.img_k_proj = nn.Linear(dim, dim)

        # Initialize the linear layers for the text
        self.text_v_proj = nn.Linear(dim, dim)
        self.text_k_proj = nn.Linear(dim, dim)
        self.q_proj = nn.Linear(dim, dim)

        # Initialize the linear layer
        self.proj = nn.Linear(dim, dim)

        # Attention
        self.attn = FlashAttention(causal=True, dropout=dropout, flash=False)

    def modality_indicator(self, x):
        """Function that returns the modality indicator"""
        if x.dim() == 4:
            return 0
        elif x.dim() == 3:
            return 1
        else:
            raise ValueError("The tensor must be 3 or 4 dimensions")

        # indicator = nn.Linear(self.dim, self.heads)
        # modality_weights = torch.sigmoid(indicator(x))
        # return modality_weights

    # def forward(self, text, img):
    #     """Forward pass of the modality adaptive module"""

    #     # Normalize the text and image features
    #     text_normalized = self.norm_text(text)
    #     img_normalized = self.norm_img(img)

    #     # Concatenate the normalized text and image features
    #     norms_concat = torch.concat((text_normalized, img_normalized))

    #     # Project the text and image features to the same dimension
    #     vision_v = self.img_v_proj(img_normalized)
    #     vision_k = self.img_k_proj(img_normalized)
    #     # Text features are projected to the same dimension as the image features
    #     text_v = self.text_v_proj(text_normalized)
    #     text_k = self.text_k_proj(text_normalized)

    #     # Combine keys from both modalities
    #     k = torch.cat((text_k, vision_k))
    #     v = torch.cat((text_v, vision_v))

    #     # # Project the query to the same dimension as the image and text features
    #     q = self.q_proj(norms_concat)

    #     # # Matmul between the query and the keys
    #     # matmuled = torch.matmul(q, keys_combined)

    #     # # add scale
    #     # matmul_scale = matmuled * self.scale

    #     # # Attention mechanism: dot product of queries and keys, scaled and normalized
    #     # attn = torch.softmax(matmul_scale)

    #     # # Matmul between the softmaxed matmuled and the values
    #     # x = torch.matmul(attn, values_combined)

    #     attn = self.attn(q, k, v)

    #     # Projected matmul
    #     x = self.proj(attn)

    #     # Normalize the outputs
    #     normed_text = self.norm_text(x)
    #     normed_img = self.norm_img(x)
    #     x = torch.concat((normed_text, normed_img))

    #     return x

    def forward(self, text, img):
        batch_size = text.size(0)

        # Normalize the text and image features
        text_normalized = self.norm_text(text)
        img_normalized = self.norm_img(img)

        # Project the text and image features to the same dimension
        vision_v = self.img_v_proj(img_normalized).view(
            batch_size, -1, self.heads, self.dim // self.heads
        )
        vision_k = self.img_k_proj(img_normalized).view(
            batch_size, -1, self.heads, self.dim // self.heads
        )
        text_v = self.text_v_proj(text_normalized).view(
            batch_size, -1, self.heads, self.dim // self.heads
        )
        text_k = self.text_k_proj(text_normalized).view(
            batch_size, -1, self.heads, self.dim // self.heads
        )

        # Combine keys and values from both modalities
        keys_combined = torch.cat((text_k, vision_k), dim=1)
        values_combined = torch.cat((text_v, vision_v), dim=1)

        # Project the query to the same dimension as the image and text features
        queries = self.q_proj(
            torch.cat((text_normalized, img_normalized), dim=1)
        )
        queries = queries.view(
            batch_size, -1, self.heads, self.dim // self.heads
        )

        # Compute the scaled dot-product attention
        # (batch_size, heads, seq_len_q, seq_len_k)
        attention_scores = torch.einsum(
            "bhid,bhjd->bhij", queries, keys_combined
        )
        attention_scores = attention_scores * self.scale
        attention_weights = F.softmax(attention_scores, dim=-1)

        # Apply the attention to the values
        # (batch_size, heads, seq_len_q, depth_v)
        attention_output = torch.einsum(
            "bhij,bhjd->bhid", attention_weights, values_combined
        )

        # Concatenate the heads
        attention_output = attention_output.contiguous().view(
            batch_size, -1, self.dim
        )

        # Apply dropout if necessary
        attention_output = F.dropout(
            attention_output, p=self.dropout, training=self.training
        )

        # Project the output of the attention mechanism
        x = self.proj(attention_output)

        # Normalize the outputs
        normed_text = self.norm_text(x)
        normed_img = self.norm_img(x)
        x = normed_text + normed_img

        return x


x = torch.randn(1, 3, 512)
y = torch.randn(1, 3, 512)

model = ModalityAdaptiveModule(512, 8)

out = model(x, y)

print(out.shape)



import functools
import logging

import torch

# Logging initialization
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)


# Main function
def track_cuda_memory_usage(func):
    """Track CUDA memory usage of a function.

    Args:
    func (function): The function to be tracked.

    Returns:
    function: The wrapped function.

    Example:
        >>> @track_cuda_memory_usage
        >>> def train():
        >>>     pass
        >>> train()
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not torch.cuda.is_available():
            logging.warning("CUDA is not available, skip tracking memory usage")
            return func(*args, **kwargs)

        torch.cuda.synchronize()
        before_memory = torch.cuda.memory_allocated()

        try:
            result = func(*args, **kwargs)
        except Exception as error:
            logging.error(f"Error occurs when running {func.__name__}: {error}")
            raise

        finally:
            torch.cuda.synchronize()
            after_memory = torch.cuda.memory_allocated()
            memory_diff = after_memory - before_memory
            logging.info(
                f"Memory usage of {func.__name__}: {memory_diff} bytes"
            )

        return result

    return wrapper



import jax.numpy as jnp
import numpy as np
import tensorflow as tf
import torch


class Matrix:
    """Matrix class that can be converted between frameworks


    Args:
        data (torch.Tensor, jnp.ndarray, tf.Tensor): Data to be converted

    Example:
    >>> import torch
    >>> import jax.numpy as jnp
    >>> import tensorflow as tf
    >>> from zeta.nn.modules.matrix import Matrix
    >>>
    >>> tensor1 = Matrix(torch.tensor([1, 2, 3]))
    >>> tensor2 = Matrix(jnp.array([1, 2, 3]))
    >>> tensor3 = Matrix(tf.constant([1, 2, 3]))
    >>>
    >>> print(tensor1.to_jax())
    >>> print(tensor2.to_pytorch())
    >>> print(tensor3.to_tensorflow())


    """

    def __init__(self, data):
        self.data = data
        self.framework = self._detect_framework(data)

    def _detect_framework(self, data):
        """Detect framework

        Args:
            data (_type_): _description_

        Raises:
            TypeError: _description_

        Returns:
            _type_: _description_
        """
        if isinstance(data, torch.Tensor):
            return "pytorch"
        elif isinstance(data, jnp.ndarray):
            return "jax"
        elif isinstance(data, tf.Tensor):
            return "tensorflow"
        else:
            raise TypeError("Unknown framework")

    def to_pytorch(self):
        """TODO: Docstring for to_pytorch.

        Returns:
            _type_: _description_
        """
        if self.framework == "pytorch":
            return self.data
        elif self.framework == "jax":
            # Convert JAX array to numpy array first, then to PyTorch tensor
            numpy_data = np.array(self.data)  # Convert JAX array to numpy array
            return torch.tensor(
                numpy_data
            )  # Convert numpy array to PyTorch tensor
        elif self.framework == "tensorflow":
            return torch.tensor(self.data.numpy())

    def to_jax(self):
        """To jax

        Returns:
            _type_: _description_
        """
        if self.framework == "jax":
            return self.data
        elif self.framework == "pytorch":
            return jnp.array(self.data.cpu().numpy())
        elif self.framework == "tensorflow":
            return jnp.array(self.data.numpy())

    def to_tensorflow(self):
        """To tensorflow

        Returns:
            _type_: _description_
        """
        if self.framework == "tensorflow":
            return self.data
        elif self.framework == "pytorch":
            return tf.convert_to_tensor(self.data.numpy.cpu().numpy())
        elif self.framework == "jax":
            return tf.convert_to_tensor(self.data)

    def sum(self):
        """Sum

        Returns:
            _type_: _description_
        """
        if self.framework == "pytorch":
            return self.data.sum()
        elif self.framework == "jax":
            return jnp.sum(self.data)
        elif self.framework == "tensorflow":
            return tf.reduce_sum(self.data)


# # Example usage
# tensor1 = Matrix(torch.tensor([1, 2, 3]))
# tensor2 = Matrix(jnp.array([1, 2, 3]))
# tensor3 = Matrix(tf.constant([1, 2, 3]))

# print(tensor1.to_jax())
# print(tensor2.to_pytorch())
# print(tensor3.to_tensorflow())



import os
import subprocess

import torch

# from setuptools import setup
from torch.utils.cpp_extension import (
    CUDA_HOME,
)  # , BuildExtension, CUDAExtension

# ninja build does not work unless include_dirs are abs path
this_dir = os.path.dirname(os.path.abspath(__file__))


def get_cuda_bare_metal_version(cuda_dir: str):
    """
    Retrieves the bare metal version of CUDA installed in the specified directory.

    Args:
        cuda_dir (str): The directory where CUDA is installed.

    Returns:
        tuple: A tuple containing the raw output of the command, the major version of the bare metal CUDA, and the minor version of the bare metal CUDA.
    """
    raw_output = subprocess.check_output(
        [cuda_dir + "/bin/nvcc", "-V"], text=True
    )
    output = raw_output.split()
    release_idx = output.index("release") + 1
    release = output[release_idx].split(".")
    bare_metal_major = release[0]
    bare_metal_minor = release[1][0]

    return raw_output, bare_metal_major, bare_metal_minor


def check_cuda_torch_binary_vs_bare_metal(cuda_dir: str):
    """
    Compares the version of CUDA used to compile PyTorch binaries with the version
    of CUDA used to compile CUDA extensions. Raises a RuntimeError if there is a
    version mismatch.

    Args:
        cuda_dir (str): The directory path where CUDA is installed.

    Raises:
        RuntimeError: If the version of CUDA used to compile CUDA extensions does
            not match the version used to compile PyTorch binaries.

    Returns:
        None
    """
    (
        raw_output,
        bare_metal_major,
        bare_metal_minor,
    ) = get_cuda_bare_metal_version(cuda_dir)
    torch_binary_major = torch.version.cuda.split(".")[0]
    torch_binary_minor = torch.version.cuda.split(".")[1]

    print("\nCompiling cuda extensions with")
    print(raw_output + "from " + cuda_dir + "/bin\n")

    if (bare_metal_major != torch_binary_major) or (
        bare_metal_minor != torch_binary_minor
    ):
        raise RuntimeError(
            "Cuda extensions are being compiled with a version of Cuda that"
            " does not match the version used to compile Pytorch binaries. "
            " Pytorch binaries were compiled with Cuda {}.\n".format(
                torch.version.cuda
            )
            + "In some cases, a minor-version mismatch will not cause later"
            " errors: "
            " https://github.com/NVIDIA/apex/pull/323#discussion_r287021798. "
            " You can try commenting out this check (at your own risk)."
        )


def raise_if_cuda_home_none(global_option: str) -> None:
    if CUDA_HOME is not None:
        return
    raise RuntimeError(
        f"{global_option} was requested, but nvcc was not found.  Are you sure"
        " your environment has nvcc available?  If you're installing within a"
        " container from https://hub.docker.com/r/pytorch/pytorch, only images"
        " whose names contain 'devel' will provide nvcc."
    )


def append_nvcc_threads(nvcc_extra_args):
    _, bare_metal_major, bare_metal_minor = get_cuda_bare_metal_version(
        CUDA_HOME
    )
    if int(bare_metal_major) >= 11 and int(bare_metal_minor) >= 2:
        return nvcc_extra_args + ["--threads", "4"]
    return nvcc_extra_args


def check_cuda():
    if not torch.cuda.is_available():
        # https://github.com/NVIDIA/apex/issues/486
        # Extension builds after https://github.com/pytorch/pytorch/pull/23408 attempt to query torch.cuda.get_device_capability(),
        # which will fail if you are compiling in an environment without visible GPUs (e.g. during an nvidia-docker build command).
        print(
            "\nWarning: Torch did not find available GPUs on this system.\n",
            (
                "If your intention is to cross-compile, this is not an"
                " error.\nBy default, Apex will cross-compile for Pascal"
                " (compute capabilities 6.0, 6.1, 6.2),\nVolta (compute"
                " capability 7.0), Turing (compute capability 7.5),\nand, if"
                " the CUDA version is >= 11.0, Ampere (compute capability"
                " 8.0).\nIf you wish to cross-compile for a single specific"
                ' architecture,\nexport TORCH_CUDA_ARCH_LIST="compute'
                ' capability" before running setup.py.\n'
            ),
        )
        if os.environ.get("TORCH_CUDA_ARCH_LIST", None) is None:
            _, bare_metal_major, bare_metal_minor = get_cuda_bare_metal_version(
                CUDA_HOME
            )
            if int(bare_metal_major) == 11:
                os.environ["TORCH_CUDA_ARCH_LIST"] = "6.0;6.1;6.2;7.0;7.5;8.0"
                if int(bare_metal_minor) > 0:
                    os.environ["TORCH_CUDA_ARCH_LIST"] = (
                        "6.0;6.1;6.2;7.0;7.5;8.0;8.6"
                    )
            else:
                os.environ["TORCH_CUDA_ARCH_LIST"] = "6.0;6.1;6.2;7.0;7.5"


# print("\n\ntorch.__version__  = {}\n\n".format(torch.__version__))
# TORCH_MAJOR = int(torch.__version__.split(".")[0])
# TORCH_MINOR = int(torch.__version__.split(".")[1])

# cmdclass = {}
# ext_modules = []

# raise_if_cuda_home_none("flashmm")
# # Check, if CUDA11 is installed for compute capability 8.0
# cc_flag = []
# # cc_flag.append("-gencode")
# # cc_flag.append("arch=compute_70,code=sm_70")
# cc_flag.append("-gencode")
# cc_flag.append("arch=compute_80,code=sm_80")

# ext_modules.append(
#     CUDAExtension(
#         'flashmm', [
#             'flash_mm.cpp',
#             'mm_block_fwd_cuda.cu',
#             'hyena_filter_cuda.cu',
#         ],
#         extra_compile_args={'cxx': ['-g', '-march=native', '-funroll-loops'],
#                             'nvcc': ['-O3', '--threads', '4', '-lineinfo', '--use_fast_math', '-std=c++17', '-arch=compute_70']
#         # extra_compile_args={'cxx': ['-O3'],
#         #                     'nvcc': append_nvcc_threads(['-O3', '-lineinfo', '--use_fast_math', '-std=c++17'] + cc_flag)
#                             },
#         include_dirs=[os.path.join(this_dir, 'mathdx/22.02/include')],
#     )
# )

# torch.utils.cpp_extension.COMMON_NVCC_FLAGS.remove('-D__CUDA_NO_HALF2_OPERATORS__')

# setup(
#     name="flashmm",
#     version="0.1",
#     description="Fast modules for Monarch Mixer block",
#     ext_modules=ext_modules,
#     cmdclass={"build_ext": BuildExtension} if ext_modules else {},
# )



import torch
import torch.nn.functional as F
from einops import pack, rearrange, unpack
from torch import nn

from zeta.utils.main import once  # noqa: F401
from zeta.utils.main import (
    eval_decorator,
    exists,
    top_a,
    top_k,
    top_p,
)  # noqa: E402


# Utils
def temperature_sampling(self, logits, temperature):
    """
    Temperature sampling.
    """
    return torch.multinomial(F.softmax(logits / temperature, dim=-1), 1)


def top_p_sampling(self, logits, p):
    """
    top-p sampling.

    Args:
        logits (torch.Tensor): The logits.
        p (float): The probability mass to keep.

    Returns:
        torch.Tensor: The sampled token.

    """
    sorted_logits, sorted_indices = torch.sort(logits, descending=True)
    cumulative_probs = torch.cumsum(F.softmax(sorted_logits, dim=-1), dim=-1)

    sorted_indices_to_remove = cumulative_probs > p
    sorted_indices_to_remove[..., 1:] = sorted_indices_to_remove[
        ..., :-1
    ].clone()
    sorted_indices_to_remove[..., 0] = 0

    indices_to_remove = sorted_indices[sorted_indices_to_remove]
    logits[indices_to_remove] = float("-inf")
    return torch.multinomial(F.softmax(logits, dim=-1), 1)


def classifier_free_guidance(self, logits_cond, logits_uncond, alpha):
    """
    Classifier-free guidance.

    Args:
        logits_cond (torch.Tensor): The conditional logits.
        logits_uncond (torch.Tensor): The unconditional logits.
        alpha (float): The alpha parameter.

    Examples::

                >>> net = nn.Linear(10, 10)
                >>> net = AutoregressiveWrapper(net)
                >>> x = torch.randn(1, 10)
                >>> logits = net(x)
                >>> print(logits.shape)
                torch.Size([1, 10, 10]) # (batch_size, seq_len, vocab_size)

    """
    return logits_uncond + alpha * (logits_cond - logits_uncond)


def contrastive_guidance(self, logits, k):
    """
    Contrastive guidance.

    Args:
        logits (torch.Tensor): The logits.
        k (int): The number of guesses to use.

    Returns:
        torch.Tensor: The sampled token.


    """
    top_k_logits, _ = torch.topk(logits, k)
    return torch.multinomial(F.softmax(top_k_logits, dim=-1), 1)


class AutoregressiveWrapper(nn.Module):
    """

    Auto-regressive wrapper for any nn.Module that takes in a sequence of
    tokens and outputs a sequence of logits.

    Args:
        net (nn.Module): A nn.Module that takes in a sequence of tokens and
            outputs a sequence of logits.
        ignore_index (int): The index to ignore in the target sequence.
        pad_value (int): The value to pad the target sequence with.
        mask_prob (float): The probability of masking out a token in the
            input sequence.
        speculative (bool): Whether to use speculative decoding or not.

    Examples::

            >>> net = nn.Linear(10, 10)
            >>> net = AutoregressiveWrapper(net)
            >>> x = torch.randn(1, 10)
            >>> logits = net(x)
            >>> print(logits.shape)
            torch.Size([1, 10, 10]) # (batch_size, seq_len, vocab_size)

    """

    def __init__(
        self,
        net,
        ignore_index=-100,
        pad_value=0,
        mask_prob=0.0,
        speculative=False,
    ):
        super().__init__()
        self.pad_value = pad_value
        self.ignore_index = ignore_index

        self.net = net
        self.max_seq_len = net.max_seq_len

        # paper shows masking (MLM) in conjunction with autoregressive
        # decoder-only training leads to big improvements
        # https://arxiv.org/abs/2210.13432
        assert mask_prob < 1.0
        self.mask_prob = mask_prob

    @torch.no_grad()
    @eval_decorator
    def generate(
        self,
        start_tokens,
        seq_len,
        eos_token=None,
        strategy="temperature",
        temperature=1.0,
        filter_logits_fn=top_k,
        filter_thres=0.9,
        min_p_pow=2.0,
        min_p_ratio=0.02,
        gamma=5,  # number of guesses for speculative decoding
        **kwargs,
    ):
        """
        Generate a sequence of tokens from the model.

        Args:
            start_tokens (torch.Tensor): The starting tokens.
            seq_len (int): The length of the sequence to generate.
            eos_token (int): The token to stop generation at.
            strategy (str): The generation strategy to use.
            temperature (float): The temperature to use for sampling.
            filter_logits_fn (function): The function to use to filter logits.
            filter_thres (float): The threshold to use for filtering logits.
            min_p_pow (float): The power to use for top-a filtering.
            min_p_ratio (float): The ratio to use for top-a filtering.
            gamma (int): The number of guesses to use for speculative decoding.
            **kwargs: Keyword arguments for the wrapped module.

        Returns:
            torch.Tensor: The generated sequence of tokens.

        Examples::

                    >>> net = nn.Linear(10, 10)
                    >>> net = AutoregressiveWrapper(net)
                    >>> x = torch.randn(1, 10)
                    >>> generated = net.generate(x, 10)
                    >>> print(generated.shape)
                    torch.Size([1, 10])
        """
        start_tokens, ps = pack([start_tokens], "* n")

        b, t = start_tokens.shape

        out = start_tokens

        if self.speculative:
            for _ in range(seq_len):
                x = out[:, -self.max_seq_len]
                logits = self.net(x, **kwargs)[:, -1]

                if filter_logits_fn in {top_k, top_p}:
                    filtered_logits = filter_logits_fn(
                        logits, thres=filter_thres
                    )
                    probs = F.softmax(filtered_logits / temperature, dim=-1)
                elif filter_logits_fn is top_a:
                    filtered_logits = filter_logits_fn(
                        logits, min_p_pow=min_p_pow, min_p_ratio=min_p_ratio
                    )
                    probs = F.softmax(filtered_logits / temperature, dim=-1)

                # speculative decoding
                guesses = torch.multinomial(probs, gamma, replacement=True)

                p_values = []
                for guess in guesses:
                    x_prime = torch.cat((x, guess.unsqueeze(0)), dim=1)
                    logits_prime = self.net(x_prime, **kwargs)[:, -1]
                    p_values.append(
                        F.softmax(logits_prime / temperature, dim=-1)
                    )

                n = gamma
                for i in range(gamma):
                    ri = torch.rand(1).item()
                    if (
                        ri
                        > p_values[i][guesses[i].item()]
                        / probs[guesses[i].item()]
                    ):
                        n = i - 1
                        break

                p_0 = p_values[n]
                if n < gamma:
                    q_n = probs[guesses[n].item()]
                    p_0 = F.normalize(torch.clamp(p_0 - q_n, min=0), p=1, dim=0)

                sample = torch.multinomial(p_0, 1)

                out = torch.cat((out, sample), dim=-1)

                if exists(eos_token):
                    is_eos_tokens = out == eos_token

                    if is_eos_tokens.any(dim=-1).all():
                        # mask out everything after the eos tokens
                        shifted_is_eos_tokens = F.pad(is_eos_tokens, (1, -1))
                        mask = shifted_is_eos_tokens.float().cumsum(dim=-1) >= 1
                        out = out.masked_fill(mask, self.pad_value)
                        break

                out = out[:, t:]
                (out,) = unpack(out, ps, "* n")
                return out
        else:
            for _ in range(seq_len):
                x = out[:, -self.max_seq_len :]

                logits = self.net(x, **kwargs)[:, -1]

                if filter_logits_fn in {top_k, top_p}:
                    filtered_logits = filter_logits_fn(
                        logits, thres=filter_thres
                    )
                    probs = F.softmax(filtered_logits / temperature, dim=-1)

                elif filter_logits_fn is top_a:
                    filtered_logits = filter_logits_fn(
                        logits, min_p_pow=min_p_pow, min_p_ratio=min_p_ratio
                    )
                    probs = F.softmax(filtered_logits / temperature, dim=-1)

                sample = torch.multinomial(probs, 1)

                out = torch.cat((out, sample), dim=-1)

                if exists(eos_token):
                    is_eos_tokens = out == eos_token

                    if is_eos_tokens.any(dim=-1).all():
                        # mask out everything after the eos tokens
                        shifted_is_eos_tokens = F.pad(is_eos_tokens, (1, -1))
                        mask = shifted_is_eos_tokens.float().cumsum(dim=-1) >= 1
                        out = out.masked_fill(mask, self.pad_value)
                        break

            out = out[:, t:]

            (out,) = unpack(out, ps, "* n")

            return out

    def forward(self, x, return_loss=True, **kwargs):
        """
        Forward pass of the autoregressive wrapper.

        Args:
            x (torch.Tensor): Input tensor.
            return_loss (bool): Whether to return the loss or not.
            **kwargs: Keyword arguments for the wrapped module.

        Returns:
            torch.Tensor: Output tensor.
            torch.Tensor: Loss tensor if return_loss is True.

        Examples::

                >>> net = nn.Linear(10, 10)
                >>> net = AutoregressiveWrapper(net)
                >>> x = torch.randn(1, 10)
                >>> logits = net(x)
                >>> print(logits.shape)
                torch.Size([1, 10, 10]) # (batch_size, seq_len, vocab_size)

        """
        seq, ignore_index = x.shape[1], self.ignore_index

        inp, target = x[:, :-1], x[:, 1:]

        if self.mask_prob > 0.0:
            rand = torch.randn(inp.shape, device=x.device)
            # first token should not be masked out
            rand[:, 0] = -torch.finfo(rand.dtype).max
            num_mask = min(int(seq * self.mask_prob), seq - 1)
            indices = rand.topk(num_mask, dim=-1).indices
            mask = ~torch.zeros_like(inp).scatter(1, indices, 1.0).bool()
            kwargs.update(self_attn_context_mask=mask)

        logits = self.net(inp, **kwargs)

        loss = F.cross_entropy(
            rearrange(logits, "b n c -> b c n"),
            target,
            ignore_index=ignore_index,
        )

        if return_loss:
            return logits, loss

        return logits

    @torch.no_grad()
    @eval_decorator
    def generate_n_solutions(self, start_tokens, n, seqlen, **kwargs):
        """Generate n solutions from the model."""
        solutions = []
        for _ in range(n):
            generated = self.generate(start_tokens, seqlen, **kwargs)
            solutions.append(generated)
        return solutions

    def evaluate_and_select_best_solution(
        self,
        solutions,
        reward_model,
    ):
        """Evaluate solutions and select the best one."""
        scores = [reward_model(solution) for solution in solutions]
        best_solution_idx = scores.index(max(scores))
        return solutions[best_solution_idx]

    def grade_solution(self, solution):
        """Grade a solution."""



import torch
from torch import nn


class StochasticSkipBlocK(nn.Module):
    """
    A module that implements stochastic skip connections in a neural network.

    Args:
        sb1 (nn.Module): The module to be skipped with a certain probability.
        p (float): The probability of skipping the module. Default is 0.5.

    Returns:
        torch.Tensor: The output tensor after applying the stochastic skip connection.
    """

    def __init__(self, sb1, p=0.5):
        super().__init__()
        self.sb1 = sb1
        self.p = p

    def forward(self, x: torch.Tensor):
        """
        Forward pass of the StochasticDepth module.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Output tensor after applying the StochasticDepth module.
        """
        if self.training and torch.rand(1).item() < self.p:
            return x  # Skip the sb1
        else:
            return self.sb1(x)



from einops import rearrange


def image_reshape(img_batch):
    """
    Reshapes an image batch

    Einstein summation notation:
        'b' = batch size
        'h' = image height
        'w' = image width
        'c' = number of channels

    Args:
        img_batch (torch.Tensor): Image batch to be reshaped

    Returns:
        torch.Tensor: Reshaped image batch

    Usage:
        >>> img_batch = torch.rand(1, 3, 224, 224)
        >>> img_batch.shape
        torch.Size([1, 3, 224, 224])
        >>> img_batch = image_reshape(img_batch)
        >>> img_batch.shape
        torch.Size([1, 224, 224, 3])

    """
    return rearrange(img_batch, "b h w c -> b c h w")


# #random
# x = torch.rand(1, 3, 224, 224)
# model = image_reshape(x)
# print(model)



import math

import torch


# Efficient implementation equivalent to the following:
def scaled_dot_product_attention(
    query,
    key,
    value,
    attn_mask=None,
    dropout_p=0.0,
    is_causal=False,
    scale=None,
) -> torch.Tensor:
    """
    Compute scaled dot product attention.

    Args:
        query (torch.Tensor): The query tensor of shape (..., L, H).
        key (torch.Tensor): The key tensor of shape (..., S, H).
        value (torch.Tensor): The value tensor of shape (..., S, D).
        attn_mask (torch.Tensor, optional): The attention mask tensor of shape (..., L, S).
        dropout_p (float, optional): The dropout probability. Default is 0.0.
        is_causal (bool, optional): Whether to use causal attention. Default is False.
        scale (float, optional): The scale factor for the attention weights. Default is None.

    Returns:
        torch.Tensor: The attention weights tensor of shape (..., L, S) multiplied by the value tensor.

    """
    # Efficient implementation equivalent to the following:
    L, S = query.size(-2), key.size(-2)
    scale_factor = 1 / math.sqrt(query.size(-1)) if scale is None else scale
    attn_bias = torch.zeros(L, S, dtype=query.dtype)
    if is_causal:
        assert attn_mask is None
        temp_mask = torch.ones(L, S, dtype=torch.bool).tril(diagonal=0)
        attn_bias.masked_fill_(temp_mask.logical_not(), float("-inf"))
        attn_bias.to(query.dtype)

    if attn_mask is not None:
        if attn_mask.dtype == torch.bool:
            attn_bias.masked_fill_(attn_mask.logical_not(), float("-inf"))
        else:
            attn_bias += attn_mask
    attn_weight = query @ key.transpose(-2, -1) * scale_factor
    attn_weight += attn_bias
    attn_weight = torch.softmax(attn_weight, dim=-1)
    attn_weight = torch.dropout(attn_weight, dropout_p, train=True)
    return attn_weight @ value



from functools import partial

import torch
from torch.distributed.fsdp import (
    BackwardPrefetch,
    FullyShardedDataParallel,
    MixedPrecision,
    ShardingStrategy,
)
from torch.distributed.fsdp.wrap import transformer_auto_wrap_policy


def fsdp(
    model: torch.nn.Module,
    auto_wrap: bool = False,
    mp: str = "fp32",
    shard_strat: str = "NO_SHARD",
    TransformerBlock=None,
):
    """
    This function wraps a given PyTorch model with the FullyShardedDataParallel (FSDP) wrapper to enable efficient data parallelism and model sharding.

    Args:
        model (torch.nn.Module): The original PyTorch model to be wrapped with FSDP.
        auto_wrap (bool, optional): If True, it enables automatic wrapping of the model's layers according to the transformer_auto_wrap_policy. Default is False.
        mp (str, optional): The mixed precision mode to be used. Can be 'bf16' for BFloat16, 'fp16' for Float16 or 'fp32' for Float32 precision. Default is 'fp32'.
        shard_strat (str, optional): The sharding strategy to be used. Can be 'SHARD_GRAD' for sharding at gradient computation, 'FULL_SHARD' for full model sharding or 'NO_SHARD' for no sharding. Default is 'NO_SHARD'.

    Raises:
        ValueError: If the provided mp (mixed precision mode) is not 'bf16', 'fp16' or 'fp32'.
        ValueError: If the provided shard_strat (sharding strategy) is not 'SHARD_GRAD', 'FULL_SHARD' or 'NO_SHARD'.

    Returns:
        torch.nn.Module: The input model wrapped with FSDP.
    """
    if auto_wrap:
        LongNet_auto_wrap_policy = partial(
            transformer_auto_wrap_policy,
            transformer_layer_cls={
                TransformerBlock,
            },
        )
    else:
        LongNet_auto_wrap_policy = None

    if mp == "bf16":
        mp_fsdp = MixedPrecision(
            param_dtype=torch.bfloat16,
            # Gradient communication precision.
            reduce_dtype=torch.bfloat16,
            # Buffer precision.
            buffer_dtype=torch.bfloat16,
        )
    elif mp == "fp16":
        mp_fsdp = MixedPrecision(
            param_dtype=torch.float16,
            # Gradient communication precision.
            reduce_dtype=torch.float16,
            # Buffer precision.
            buffer_dtype=torch.float16,
        )
    elif mp == "fp32":
        mp_fsdp = MixedPrecision(
            param_dtype=torch.float32,
            # Gradient communication precision.
            reduce_dtype=torch.float32,
            # Buffer precision.
            buffer_dtype=torch.float32,
        )
    else:
        raise ValueError(
            "Invalid scheduler_type. Expected 'bf16', 'fp16' or 'fp32', got:"
            f" {mp}"
        )

    if shard_strat == "SHARD_GRAD":
        sharding_strat_fsdp = ShardingStrategy.SHARD_GRAD_OP
    elif shard_strat == "FULL_SHARD":
        sharding_strat_fsdp = ShardingStrategy.FULL_SHARD
    elif shard_strat == "NO_SHARD":
        sharding_strat_fsdp = ShardingStrategy.NO_SHARD
    else:
        raise ValueError(
            "Invalid scheduler_type. Expected 'SHARD_GRAD', 'FULL_SHARD' or"
            " 'NO_SHARD', got: {}".format(shard_strat)
        )

    model = FullyShardedDataParallel(
        model,
        auto_wrap_policy=LongNet_auto_wrap_policy,
        mixed_precision=mp_fsdp,
        backward_prefetch=BackwardPrefetch.BACKWARD_PRE,
        sharding_strategy=sharding_strat_fsdp,
        forward_prefetch=True,
        use_orig_params=True,
    )

    return model



# QKV Normalization

from torch import nn


def qkv_norm(
    q,
    k,
    v,
):
    """Apply QKV normalization.

    Args:
        q (torch.Tensor): Query tensor.
        k (torch.Tensor): Key tensor.
        v (torch.Tensor): Value tensor.

    Returns:
        torch.Tensor: Normalized query, key, and value tensors.
    """
    q = nn.LayerNorm(q.size())(q)
    k = nn.LayerNorm(k.size())(k)
    v = nn.LayerNorm(v.size())(v)
    return q, k, v


def qk_norm(
    q,
    k,
):
    """Apply QK normalization.

    Args:
        q (torch.Tensor): Query tensor.
        k (torch.Tensor): Key tensor.
        v (torch.Tensor): Value tensor.

    Returns:
        torch.Tensor: Normalized query, key, and value tensors.
    """
    q = nn.LayerNorm(q.size())(q)
    k = nn.LayerNorm(k.size())(k)
    return q, k



from zeta.structs.auto_regressive_wrapper import AutoregressiveWrapper
from zeta.structs.clip_encoder import CLIPVisionTower, build_vision_tower
from zeta.structs.encoder_decoder import EncoderDecoder
from zeta.structs.hierarchical_transformer import (
    HierarchicalBlock,
    HierarchicalTransformer,
)
from zeta.structs.local_transformer import LocalTransformer
from zeta.structs.multi_modal_projector import build_vision_projector
from zeta.structs.simple_transformer import (
    ParallelTransformerBlock,
    SimpleTransformer,
)
from zeta.structs.simple_vision_encoder import VisionEncoder
from zeta.structs.transformer import (
    Decoder,
    Encoder,
    Transformer,
    ViTransformerWrapper,
)
from zeta.structs.transformer_block import TransformerBlock

__all__ = [
    "AutoregressiveWrapper",
    "Encoder",
    "Decoder",
    "EncoderDecoder",
    "HierarchicalBlock",
    "HierarchicalTransformer",
    "LocalTransformer",
    "ParallelTransformerBlock",
    "Transformer",
    "TransformerBlock",
    "ViTransformerWrapper",
    "VideoTokenizer",
    "ParallelTransformerBlock",
    "SimpleTransformer",
    "CLIPVisionTower",
    "build_vision_tower",
    "build_vision_projector",
    "VisionEncoder",
]



from zeta.utils.disable_logging import disable_warnings_and_logs

disable_warnings_and_logs()

from zeta.cloud import *  # noqa: F403, E402
from zeta.models import *  # noqa: F403, E402
from zeta.nn import *  # noqa: F403, E402
from zeta.ops import *  # noqa: F403, E402
from zeta.optim import *  # noqa: F403, E402
from zeta.quant import *  # noqa: F403, E402
from zeta.rl import *  # noqa: F403, E402
from zeta.tokenizers import *  # noqa: F403, E402
from zeta.training import *  # noqa: F403, E402
from zeta.utils import *  # noqa: F403, E402



from typing import List

import torch
from torch import Tensor, nn


class MultiModalEmbedding(nn.Module):
    """
    MultiModalEmbedding class represents a module for multi-modal embedding.

    Args:
        video_dim (int): The dimension of the video input.
        text_dim (int): The dimension of the text input.

    Attributes:
        video_embedding (nn.Linear): Linear layer for video embedding.
        text_embedding (nn.EmbeddingBag): Embedding layer for text embedding.

    Methods:
        forward(video, text): Performs forward pass of the multi-modal embedding.

    Returns:
        torch.Tensor: Concatenated tensor of video and text embeddings.
    """

    def __init__(self, video_dim, text_dim):
        super().__init__()
        self.video_embedding = nn.Linear(video_dim, 512)
        self.text_embedding = nn.EmbeddingBag(text_dim, 512, sparse=True)

    def forward(self, video, text):
        video_embed = self.video_embedding(video)
        text_embed = self.text_embedding(text)
        return torch.cat([video_embed, text_embed], dim=-1)


class MultiInputMultiModalConcatenation(nn.Module):
    """
    A module that concatenates multiple input tensors along a specified dimension.

    Args:
        dim (int): The dimension along which the input tensors will be concatenated.

    Attributes:
        dim (int): The dimension along which the input tensors will be concatenated.
    """

    def __init__(self, dim: int, *args, **kwargs):
        super().__init__()
        self.dim = dim

    def forward(self, inputs: List[Tensor]):
        """
        Forward pass of the module.

        Args:
            inputs (List[Tensor]): A list of input tensors to be concatenated.

        Returns:
            Tensor: The concatenated tensor.
        """
        return torch.cat(inputs, dim=self.dim)


class SplitMultiOutput(nn.Module):
    """
    Splits the input tensor into multiple outputs along a specified dimension.

    Args:
        dim (int): The dimension along which to split the input tensor.
        num_splits (int): The number of splits to create.
        output_dims (List[int]): The sizes of the output tensors along the split dimension.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Attributes:
        dim (int): The dimension along which to split the input tensor.
        num_splits (int): The number of splits to create.
        output_dims (List[int]): The sizes of the output tensors along the split dimension.
    """

    def __init__(
        self,
        dim: int,
        num_splits: int,
        output_dims: List[int],
        *args,
        **kwargs,
    ):
        super().__init__()
        self.dim = dim
        self.num_splits = num_splits
        self.output_dims = output_dims

    def forward(self, x: Tensor):
        """
        Forward pass of the SplitMultiOutput module.

        Args:
            x (Tensor): The input tensor to be split.

        Returns:
            Tuple[Tensor]: A tuple of output tensors after splitting the input tensor.
        """
        return torch.split(x, self.output_dims, dim=self.dim)


class OutputHead(nn.Module):
    def __init__(self, dim: int, dim_range: int, *args, **kwargs):
        """
        Initializes an OutputHead module.

        Args:
            dim (int): The input dimension.
            dim_range (int): The dimension range for softmax operation.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        super().__init__()
        self.dim = dim
        self.dim_range = dim_range

        # Linear layer for each output
        self.output_layers = nn.Sequential(
            nn.LayerNorm(dim),
            nn.Linear(dim, dim),
            nn.Softmax(dim_range),
        )

    def forward(self, x: Tensor):
        """
        Forward pass of the OutputHead module.

        Args:
            x (Tensor): The input tensor.

        Returns:
            Tensor: The output tensor.
        """
        return self.output_layers(x)


class DynamicOutputDecoder(nn.Module):
    """
    Decoder module for dynamic output.

    Args:
        input_dim (int): The input dimension.
        robot_count (int): The number of robots.

    Attributes:
        decoders (nn.ModuleList): List of linear decoders.

    """

    def __init__(self, input_dim, robot_count):
        super().__init__()
        self.decoders = nn.ModuleList(
            [nn.Linear(input_dim, input_dim) for _ in range(robot_count)]
        )

    def forward(self, x):
        """
        Forward pass of the decoder.

        Args:
            x (torch.Tensor): The input tensor.

        Returns:
            List[torch.Tensor]: List of decoded tensors.

        """
        return [decoder(x) for decoder in self.decoders]


class DynamicInputChannels(nn.Module):
    """
    A module that applies linear transformations to input data for multiple robots.

    Args:
        num_robots (int): The number of robots.
        input_dim (int): The input dimension.
        output_dim (int): The output dimension.

    Attributes:
        layers (nn.ModuleList): A list of linear layers.

    Methods:
        forward(x): Forward pass of the module.

    """

    def __init__(self, num_robots, input_dim, output_dim):
        super().__init__()
        self.layers = nn.ModuleList(
            [nn.Linear(input_dim, output_dim) for _ in range(num_robots)]
        )

    def forward(self, x):
        outputs = [layer(x) for layer in self.layers]
        return torch.cat(outputs, dim=1)


class OutputDecoders(nn.Module):
    """
    Class representing the output decoders for multiple robots.

    Args:
        num_robots (int): The number of robots.
        input_dim (int): The input dimension.
        output_dim (int): The output dimension.

    Attributes:
        decoders (nn.ModuleList): List of linear decoders for each robot.

    Methods:
        forward(x): Forward pass of the decoders.

    """

    def __init__(self, num_robots, input_dim, output_dim):
        super().__init__()
        self.decoders = nn.ModuleList(
            [nn.Linear(input_dim, output_dim) for _ in range(num_robots)]
        )

    def forward(self, x):
        """
        Forward pass of the decoders.

        Args:
            x (torch.Tensor): Input tensor.

        Returns:
            torch.Tensor: Stacked output tensor from each decoder.

        """
        return torch.stack([decoder(x) for decoder in self.decoders], dim=1)



from functools import partial
from typing import List, Union

import torch
import torch.nn.functional as F
from einops import rearrange, repeat
from torch import Tensor, nn
from torch.nn.utils.rnn import pad_sequence as orig_pad_sequence

# helpers


def exists(val):
    return val is not None


def default(val, d):
    return val if exists(val) else d


def always(val):
    return lambda *args: val


def pair(t):
    return t if isinstance(t, tuple) else (t, t)


def divisible_by(numer, denom):
    return (numer % denom) == 0


def group_images_by_max_seq_len(
    images: List[Tensor],
    patch_size: int,
    calc_token_dropout=None,
    max_seq_len=2048,
) -> List[List[Tensor]]:
    calc_token_dropout = default(calc_token_dropout, always(0.0))

    groups = []
    group = []
    seq_len = 0

    if isinstance(calc_token_dropout, (float, int)):
        calc_token_dropout = always(calc_token_dropout)

    for image in images:
        assert isinstance(image, Tensor)

        image_dims = image.shape[-2:]
        ph, pw = map(lambda t: t // patch_size, image_dims)

        image_seq_len = ph * pw
        image_seq_len = int(
            image_seq_len * (1 - calc_token_dropout(*image_dims))
        )

        assert (
            image_seq_len <= max_seq_len
        ), f"image with dimensions {image_dims} exceeds maximum sequence length"

        if (seq_len + image_seq_len) > max_seq_len:
            groups.append(group)
            group = []
            seq_len = 0

        group.append(image)
        seq_len += image_seq_len

    if len(group) > 0:
        groups.append(group)

    return groups


# normalization
# they use layernorm without bias, something that pytorch does not offer
class LayerNorm(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.gamma = nn.Parameter(torch.ones(dim))
        self.register_buffer("beta", torch.zeros(dim))

    def forward(self, x):
        return F.layer_norm(x, x.shape[-1:], self.gamma, self.beta)


# they use a query-key normalization that is equivalent to rms norm (no mean-centering, learned gamma), from vit 22B paper
class RMSNorm(nn.Module):
    def __init__(self, heads, dim):
        super().__init__()
        self.scale = dim**0.5
        self.gamma = nn.Parameter(torch.ones(heads, 1, dim))

    def forward(self, x):
        normed = F.normalize(x, dim=-1)
        return normed * self.scale * self.gamma


# feedforward
def FeedForward(dim, hidden_dim, dropout=0.0):
    return nn.Sequential(
        LayerNorm(dim),
        nn.Linear(dim, hidden_dim),
        nn.GELU(),
        nn.Dropout(dropout),
        nn.Linear(hidden_dim, dim),
        nn.Dropout(dropout),
    )


# attention
class Attention(nn.Module):
    def __init__(self, dim, heads=8, dim_head=64, dropout=0.0):
        super().__init__()
        inner_dim = dim_head * heads
        self.heads = heads
        self.norm = LayerNorm(dim)

        self.q_norm = RMSNorm(heads, dim_head)
        self.k_norm = RMSNorm(heads, dim_head)

        self.attend = nn.Softmax(dim=-1)
        self.dropout = nn.Dropout(dropout)

        self.to_q = nn.Linear(dim, inner_dim, bias=False)
        self.to_kv = nn.Linear(dim, inner_dim * 2, bias=False)

        self.to_out = nn.Sequential(
            nn.Linear(inner_dim, dim, bias=False), nn.Dropout(dropout)
        )

    def forward(self, x, context=None, mask=None, attn_mask=None):
        x = self.norm(x)
        kv_input = default(context, x)

        qkv = (self.to_q(x), *self.to_kv(kv_input).chunk(2, dim=-1))

        q, k, v = map(
            lambda t: rearrange(t, "b n (h d) -> b h n d", h=self.heads), qkv
        )

        q = self.q_norm(q)
        k = self.k_norm(k)

        dots = torch.matmul(q, k.transpose(-1, -2))

        if exists(mask):
            mask = rearrange(mask, "b j -> b 1 1 j")
            dots = dots.masked_fill(~mask, -torch.finfo(dots.dtype).max)

        if exists(attn_mask):
            dots = dots.masked_fill(~attn_mask, -torch.finfo(dots.dtype).max)

        attn = self.attend(dots)
        attn = self.dropout(attn)

        out = torch.matmul(attn, v)
        out = rearrange(out, "b h n d -> b n (h d)")
        return self.to_out(out)


# transformer block
class Transformer(nn.Module):
    def __init__(self, dim, depth, heads, dim_head, mlp_dim, dropout=0.0):
        super().__init__()
        self.layers = nn.ModuleList([])
        for _ in range(depth):
            self.layers.append(
                nn.ModuleList(
                    [
                        Attention(
                            dim, heads=heads, dim_head=dim_head, dropout=dropout
                        ),
                        FeedForward(dim, mlp_dim, dropout=dropout),
                    ]
                )
            )

        self.norm = LayerNorm(dim)

    def forward(self, x, mask=None, attn_mask=None):
        for attn, ff in self.layers:
            x = attn(x, mask=mask, attn_mask=attn_mask) + x
            x = ff(x) + x

        return self.norm(x)


class NaViT(nn.Module):
    def __init__(
        self,
        *,
        image_size,
        patch_size,
        num_classes,
        dim,
        depth,
        heads,
        mlp_dim,
        channels=3,
        dim_head=64,
        dropout=0.0,
        emb_dropout=0.0,
        token_dropout_prob=None,
    ):
        super().__init__()
        image_height, image_width = pair(image_size)

        # what percent of tokens to dropout
        # if int or float given, then assume constant dropout prob
        # otherwise accept a callback that in turn calculates dropout prob from height and width

        self.calc_token_dropout = None

        if callable(token_dropout_prob):
            self.calc_token_dropout = token_dropout_prob

        elif isinstance(token_dropout_prob, (float, int)):
            assert 0.0 < token_dropout_prob < 1.0
            token_dropout_prob = float(token_dropout_prob)
            self.calc_token_dropout = lambda height, width: token_dropout_prob

        # calculate patching related stuff

        assert divisible_by(image_height, patch_size) and divisible_by(
            image_width, patch_size
        ), "Image dimensions must be divisible by the patch size."

        patch_height_dim, patch_width_dim = (image_height // patch_size), (
            image_width // patch_size
        )
        patch_dim = channels * (patch_size**2)

        self.channels = channels
        self.patch_size = patch_size

        self.to_patch_embedding = nn.Sequential(
            LayerNorm(patch_dim),
            nn.Linear(patch_dim, dim),
            LayerNorm(dim),
        )

        self.pos_embed_height = nn.Parameter(torch.randn(patch_height_dim, dim))
        self.pos_embed_width = nn.Parameter(torch.randn(patch_width_dim, dim))

        self.dropout = nn.Dropout(emb_dropout)

        self.transformer = Transformer(
            dim, depth, heads, dim_head, mlp_dim, dropout
        )

        # final attention pooling queries

        self.attn_pool_queries = nn.Parameter(torch.randn(dim))
        self.attn_pool = Attention(dim=dim, dim_head=dim_head, heads=heads)

        # output to logits

        self.to_latent = nn.Identity()

        self.mlp_head = nn.Sequential(
            LayerNorm(dim), nn.Linear(dim, num_classes, bias=False)
        )

    @property
    def device(self):
        return next(self.parameters()).device

    def forward(
        self,
        batched_images: Union[
            List[Tensor], List[List[Tensor]]
        ],  # assume different resolution images already grouped correctly
        group_images=False,
        group_max_seq_len=2048,
    ):
        p, c, device, has_token_dropout = (
            self.patch_size,
            self.channels,
            self.device,
            exists(self.calc_token_dropout),
        )

        arange = partial(torch.arange, device=device)
        pad_sequence = partial(orig_pad_sequence, batch_first=True)

        # auto pack if specified

        if group_images:
            batched_images = group_images_by_max_seq_len(
                batched_images,
                patch_size=self.patch_size,
                calc_token_dropout=self.calc_token_dropout,
                max_seq_len=group_max_seq_len,
            )

        # process images into variable lengthed sequences with attention mask

        num_images = []
        batched_sequences = []
        batched_positions = []
        batched_image_ids = []

        for images in batched_images:
            num_images.append(len(images))

            sequences = []
            positions = []
            image_ids = torch.empty((0,), device=device, dtype=torch.long)

            for image_id, image in enumerate(images):
                assert image.ndim == 3
                assert image.shape[0] == c
                image_dims = image.shape[-2:]
                assert all([divisible_by(dim, p) for dim in image_dims]), (
                    f"height and width {image_dims} of images must be divisible"
                    f" by patch size {p}"
                )

                ph, pw = map(lambda dim: dim // p, image_dims)

                pos = torch.stack(
                    torch.meshgrid((arange(ph), arange(pw)), indexing="ij"),
                    dim=-1,
                )

                pos = rearrange(pos, "h w c -> (h w) c")
                seq = rearrange(
                    image, "c (h p1) (w p2) -> (h w) (c p1 p2)", p1=p, p2=p
                )

                seq_len = seq.shape[-2]

                if has_token_dropout:
                    token_dropout = self.calc_token_dropout(*image_dims)
                    num_keep = max(1, int(seq_len * (1 - token_dropout)))
                    keep_indices = (
                        torch.randn((seq_len,), device=device)
                        .topk(num_keep, dim=-1)
                        .indices
                    )

                    seq = seq[keep_indices]
                    pos = pos[keep_indices]

                image_ids = F.pad(image_ids, (0, seq.shape[-2]), value=image_id)
                sequences.append(seq)
                positions.append(pos)

            batched_image_ids.append(image_ids)
            batched_sequences.append(torch.cat(sequences, dim=0))
            batched_positions.append(torch.cat(positions, dim=0))

        # derive key padding mask

        lengths = torch.tensor(
            [seq.shape[-2] for seq in batched_sequences],
            device=device,
            dtype=torch.long,
        )
        max_length = arange(lengths.amax().item())
        key_pad_mask = rearrange(lengths, "b -> b 1") <= rearrange(
            max_length, "n -> 1 n"
        )

        # derive attention mask, and combine with key padding mask from above

        batched_image_ids = pad_sequence(batched_image_ids)
        attn_mask = rearrange(batched_image_ids, "b i -> b 1 i 1") == rearrange(
            batched_image_ids, "b j -> b 1 1 j"
        )
        attn_mask = attn_mask & rearrange(key_pad_mask, "b j -> b 1 1 j")

        # combine patched images as well as the patched width / height positions for 2d positional embedding

        patches = pad_sequence(batched_sequences)
        patch_positions = pad_sequence(batched_positions)

        # need to know how many images for final attention pooling

        num_images = torch.tensor(num_images, device=device, dtype=torch.long)

        # to patches

        x = self.to_patch_embedding(patches)

        # factorized 2d absolute positional embedding

        h_indices, w_indices = patch_positions.unbind(dim=-1)

        h_pos = self.pos_embed_height[h_indices]
        w_pos = self.pos_embed_width[w_indices]

        x = x + h_pos + w_pos

        # embed dropout

        x = self.dropout(x)

        # attention

        x = self.transformer(x, attn_mask=attn_mask)

        # do attention pooling at the end

        max_queries = num_images.amax().item()

        queries = repeat(
            self.attn_pool_queries, "d -> b n d", n=max_queries, b=x.shape[0]
        )

        # attention pool mask

        image_id_arange = arange(max_queries)

        attn_pool_mask = rearrange(image_id_arange, "i -> i 1") == rearrange(
            batched_image_ids, "b j -> b 1 j"
        )

        attn_pool_mask = attn_pool_mask & rearrange(
            key_pad_mask, "b j -> b 1 j"
        )

        attn_pool_mask = rearrange(attn_pool_mask, "b i j -> b 1 i j")

        # attention pool

        x = (
            self.attn_pool(queries, context=x, attn_mask=attn_pool_mask)
            + queries
        )

        x = rearrange(x, "b n d -> (b n) d")

        # each batch element may not have same amount of images

        is_images = image_id_arange < rearrange(num_images, "b -> b 1")
        is_images = rearrange(is_images, "b n -> (b n)")

        x = x[is_images]

        # project out to logits

        x = self.to_latent(x)

        return self.mlp_head(x)



import torch
from torch import Tensor, nn

from zeta.structs.auto_regressive_wrapper import AutoregressiveWrapper
from zeta.structs.transformer import (
    Decoder,
    Encoder,
    Transformer,
    ViTransformerWrapper,
)


class GPT4(nn.Module):
    """
    GPT4 is a transformer-based model architecture. It initializes with
    a Transformer and AutoregressiveWrapper with default or user-specified parameters.
        Initialize the model with specified or default parameters.
        Args:
        - num_tokens: Number of tokens in the vocabulary
        - max_seq_len: Maximum sequence length
        - dim: Dimension of the model
        - depth: Depth of the model
        - dim_head: Dimension of the model head
        - heads: Number of heads
        - use_abs_pos_emb: Whether to use absolute position embedding
        - alibi_pos_bias: Alibi position bias
        - alibi_num_heads: Number of alibi heads
        - rotary_xpos: Rotary position
        - attn_flash: Attention flash
        - deepnorm: Deep normalization
        - shift_tokens: Number of tokens to shift
        - attn_one_kv_head: Attention one key/value head
        - qk_norm: Query-key normalization
        - attn_qk_norm: Attention query-key normalization
        - attn_qk_norm_dim_scale: Attention query-key normalization dimension scale
        - embedding_provider: Embedding provider module
    """

    def __init__(
        self,
        num_tokens=50432,
        max_seq_len=8192,
        dim=2560,
        depth=32,
        dim_head=128,
        heads=24,
        use_abs_pos_emb=False,
        alibi_pos_bias=True,
        alibi_num_heads=12,
        rotary_xpos=True,
        attn_flash=True,
        attn_one_kv_head=True,  # multiquery attention
        qk_norm=True,
        attn_qk_norm=True,
        attn_qk_norm_dim_scale=True,
        *args,
        **kwargs,
    ):
        super().__init__()

        try:
            self.decoder = Transformer(
                num_tokens=num_tokens,
                max_seq_len=max_seq_len,
                use_abs_pos_emb=use_abs_pos_emb,
                attn_layers=Decoder(
                    dim=dim,
                    depth=depth,
                    dim_head=dim_head,
                    heads=heads,
                    alibi_pos_bias=alibi_pos_bias,
                    alibi_num_heads=alibi_num_heads,
                    rotary_xpos=rotary_xpos,
                    attn_flash=attn_flash,
                    attn_one_kv_head=attn_one_kv_head,
                    qk_norm=qk_norm,
                    attn_qk_norm=attn_qk_norm,
                    attn_qk_norm_dim_scale=attn_qk_norm_dim_scale,
                    *args,
                    **kwargs,
                ),
            )

            self.decoder = AutoregressiveWrapper(self.decoder)

        except Exception as e:
            print("Failed to initialize Andromeda: ", e)
            raise

    def forward(self, text: Tensor, **kwargs):
        try:
            model_input = self.decoder.forward(text)[0]
            return self.decoder(model_input, padded_x=model_input[0])
        except Exception as e:
            print("Failed in forward method: ", e)
            raise


class GPT4MultiModal(torch.nn.Module):
    """
    GPT4MultiModal is a multi-modal transformer model that combines image and text inputs.

    Args:
        image_size (int): The size of the input image (default: 256).
        patch_size (int): The size of each image patch (default: 32).
        encoder_dim (int): The dimension of the encoder layers (default: 512).
        encoder_depth (int): The number of encoder layers (default: 6).
        encoder_heads (int): The number of attention heads in the encoder (default: 8).
        num_tokens (int): The number of tokens in the vocabulary (default: 20000).
        max_seq_len (int): The maximum sequence length for the decoder (default: 1024).
        decoder_dim (int): The dimension of the decoder layers (default: 512).
        decoder_depth (int): The number of decoder layers (default: 6).
        decoder_heads (int): The number of attention heads in the decoder (default: 8).
        alibi_num_heads (int): The number of attention heads for the alibi mechanism (default: 4).
        use_abs_pos_emb (bool): Whether to use absolute positional embeddings (default: False).
        cross_attend (bool): Whether to enable cross-attention between encoder and decoder (default: True).
        alibi_pos_bias (bool): Whether to use positional bias for the alibi mechanism (default: True).
        rotary_xpos (bool): Whether to use rotary positional embeddings (default: True).
        attn_flash (bool): Whether to use attention flash (default: True).
        qk_norm (bool): Whether to normalize the query-key dot product (default: True).
    """

    def __init__(
        self,
        image_size=256,
        patch_size=32,
        encoder_dim=512,
        encoder_depth=6,
        encoder_heads=8,
        num_tokens=20000,
        max_seq_len=1024,
        decoder_dim=512,
        decoder_depth=6,
        decoder_heads=8,
        alibi_num_heads=4,
        use_abs_pos_emb=False,
        cross_attend=True,
        alibi_pos_bias=True,
        rotary_xpos=True,
        attn_flash=True,
        qk_norm=True,
        *args,
        **kwargs,
    ):
        super().__init__()

        # Encoder
        self.encoder = ViTransformerWrapper(
            image_size=image_size,
            patch_size=patch_size,
            attn_layers=Encoder(
                dim=encoder_dim, depth=encoder_depth, heads=encoder_heads
            ),
        )

        # Decoder
        self.decoder = Transformer(
            num_tokens=num_tokens,
            max_seq_len=max_seq_len,
            use_abs_pos_emb=use_abs_pos_emb,
            attn_layers=Decoder(
                dim=decoder_dim,
                depth=decoder_depth,
                heads=decoder_heads,
                cross_attend=cross_attend,
                alibi_pos_bias=alibi_pos_bias,
                alibi_num_heads=alibi_num_heads,
                rotary_xpos=rotary_xpos,
                attn_flash=attn_flash,
                qk_norm=qk_norm,
            ),
        )

    def forward(self, img: Tensor, text: Tensor):
        """
        Performs the forward pass of the GPT4 model.

        Args:
            img (Tensor): The input image tensor.
            text (Tensor): The input text tensor.

        Returns:
            Tensor: The output tensor of the model.
        """
        try:
            encoded = self.encoder(img, return_embeddings=True)
            return self.decoder(text, context=encoded)
        except Exception as error:
            print(f"Failed in forward method: {error}")
            raise



import torch
import torch.nn.functional as F
from einops import rearrange
from torch import nn


class PositionalEmbedding(nn.Embedding):
    """PositionalEmbedding module.


    Args:
        d_model (int): Dimension of the model.
        max_len (int): Maximum length of the input sequence.
        padding_idx (int, optional): Index of the padding token. Defaults to 0.
        scale_grad_by_freq (bool, optional): If True, scale gradients by frequency. Defaults to False.
        sparse (bool, optional): If True, use sparse gradient updates. Defaults to False.

    Example:
        >>> positional_embedding = PositionalEmbedding(512, 1000)
        >>> x = torch.randn(32, 100, 512)
        >>> positions = torch.arange(100)
        >>> embedded_tensor = positional_embedding(x, positions)
    """

    def forward(
        self,
        x,
        positions=None,
        **kwargs,
    ):
        """
        Forward pass of the PositionalEmbedding module.

        Args:
            x (torch.Tensor): Input tensor.
            positions (torch.Tensor, optional): Positions tensor. If None, positions are generated based on the input tensor size. Default is None.
            **kwargs: Additional keyword arguments.

        Returns:
            torch.Tensor: Embedded tensor.

        """
        if positions is None:
            # being consistent with Fairseq, which starts from 2.
            positions = (
                torch.arange(2, x.size(1) + 2, device=x.device)
                .long()
                .unsqueeze(0)
            )

        positions = rearrange(positions, "b l -> l b")
        x = rearrange(x, "b l d -> l b d")
        embedded_tensor = F.embedding(
            positions,
            self.weight,
            self.padding_idx,
            self.max_norm,
            self.norm_type,
            self.scale_grad_by_freq,
            self.sparse,
        )
        embedded_tensor = rearrange(embedded_tensor, "l b d -> b l d")

        return embedded_tensor



from torch import Tensor, nn


class FusedProjSoftmax(nn.Module):
    """
    FusedProjSoftmax is a module that applies a linear projection followed by a softmax operation.

    Args:
        dim (int): The input dimension.
        dim_out (int): The output dimension.
        dim_axis (int, optional): The axis along which the softmax operation is applied. Defaults to -1.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Attributes:
        proj (nn.Linear): The linear projection layer.
        softmax (nn.Softmax): The softmax operation layer.

    Examples:
        x = torch.rand(1, 2, 3)
        model = FusedProjSoftmax(3, 4)
        out = model(x)
        print(out.shape)
    """

    def __init__(
        self, dim: int, dim_out: int, dim_axis: int = -1, *args, **kwargs
    ):
        super().__init__()
        self.proj = nn.Linear(dim, dim_out, *args, **kwargs)
        self.softmax = nn.Softmax(dim=dim_axis)

    def forward(self, x: Tensor) -> Tensor:
        """
        Forward pass of the FusedProjSoftmax module.

        Args:
            x (Tensor): The input tensor.

        Returns:
            Tensor: The output tensor after applying linear projection and softmax.
        """
        return self.softmax(self.proj(x))



import torch

from zeta.structs.auto_regressive_wrapper import AutoregressiveWrapper
from zeta.structs.transformer import (
    Decoder,
    Encoder,
    Transformer,
    ViTransformerWrapper,
)


class PalmE(torch.nn.Module):
    def __init__(
        self,
        image_size=256,
        patch_size=32,
        encoder_dim=512,
        encoder_depth=6,
        encoder_heads=8,
        num_tokens=20000,
        max_seq_len=1024,
        decoder_dim=512,
        decoder_depth=6,
        decoder_heads=8,
        alibi_num_heads=4,
        use_abs_pos_emb=False,
        cross_attend=True,
        alibi_pos_bias=True,
        rotary_xpos=True,
        attn_flash=True,
        qk_norm=True,
    ):
        super().__init__()

        self.encoder = ViTransformerWrapper(
            image_size=image_size,
            patch_size=patch_size,
            attn_layers=Encoder(
                dim=encoder_dim, depth=encoder_depth, heads=encoder_heads
            ),
        )

        self.decoder = Transformer(
            num_tokens=num_tokens,
            max_seq_len=max_seq_len,
            use_abs_pos_emb=use_abs_pos_emb,
            attn_layers=Decoder(
                dim=decoder_dim,
                depth=decoder_depth,
                heads=decoder_heads,
                cross_attend=cross_attend,
                alibi_pos_bias=alibi_pos_bias,
                alibi_num_heads=alibi_num_heads,
                rotary_xpos=rotary_xpos,
                attn_flash=attn_flash,
                qk_norm=qk_norm,
            ),
        )

        self.decoder = AutoregressiveWrapper(self.decoder)

    def forward(self, img, text):
        try:
            encoded = self.encoder(img, return_embeddings=True)
            return self.decoder(text, context=encoded)
        except Exception as error:
            print(f"Failed in forward method: {error}")
            raise



from typing import Optional

import torch
from torch import Tensor, nn

from zeta.nn.attention.cross_attention import CrossAttention
from zeta.nn.attention.multiquery_attention import MultiQueryAttention


class PerceiverLayer(nn.Module):
    """
    Perceiver Layer, this layer has a self attn that takes in q then ->
    sends the output into the q of the cross attention where the cross attn
    takes in k and v. The output of the cross attn is then sent into a
    feed forward layer.


    Args:
        dim: dimension of the input tensor
        heads: number of heads
        depth: number of layers
        dim_head: dimension of each head
        dropout: dropout rate
        ff_dropout: feed forward dropout rate
        ff_mult: feed forward multiplier

    Examples::
        >>> q = torch.randn(1, 32, 512)
        >>> k = torch.randn(1, 32, 512)
        >>> v = torch.randn(1, 32, 512)
        >>> layer = PerceiverLayer(512, 8, 6, 64)
        >>> print(layer(q, k, v).shape)
        torch.Size([1, 32, 512])

    """

    def __init__(
        self,
        dim: int,
        heads: int,
        depth: int,
        dim_head: int = 64,
        dropout: float = 0.1,
        ff_dropout: float = 0.1,
        ff_mult: int = 4,
    ):
        super().__init__()
        self.dim = dim
        self.heads = heads
        self.depth = depth
        self.dim_head = dim_head
        self.dropout = dropout
        self.ff_dropout = ff_dropout
        self.ff_mult = ff_mult

        # Initialize layers for MultiQueryAttention, CrossAttention, and Feed Forward
        self.self_attn = MultiQueryAttention(
            dim,
            heads,
            # qk_ln=True,
        )

        # CrossAttention initialization
        self.cross_attn = CrossAttention(
            dim,
            context_dim=dim,
            dim_head=dim_head,
            heads=heads,
            dropout=dropout,
        )

        # Feed Forward initialization
        self.ffn = nn.Sequential(
            nn.Linear(dim, dim * ff_mult),
            nn.GELU(),
            nn.Dropout(ff_dropout),
            nn.Linear(dim * ff_mult, dim),
            nn.Dropout(ff_dropout),
        )

        # Projection layers for x to -> q, k, v
        self.q_proj = nn.Linear(dim, dim)
        self.k_proj = nn.Linear(dim, dim)
        self.v_proj = nn.Linear(dim, dim)

    def forward(
        self,
        q: Tensor,
        k: Tensor,
        v: Tensor,
        mask: Optional[Tensor] = None,
    ):
        """
        Args:
            q: query tensor
            k: key tensor
            v: value tensor
            mask: mask tensor

        Shape:
            q: (batch_size, seq_len_q, dim)
            k: (batch_size, seq_len_k, dim)
            v: (batch_size, seq_len_v, dim)
            mask: (batch_size, seq_len_q, seq_len_k)
        """
        q, _, _ = self.self_attn(q)

        # Concatenate k and v
        kv = torch.concat((k, v), dim=1)

        # Send q, k, v into cross attention with q as the context
        x = self.cross_attn(kv, q)

        # Apply feed forward layer to output of cross attention
        x = self.ffn(x)

        # Return output
        return x



import torch
import torch.nn.functional as F
from einops import rearrange
from torch import einsum, nn


# helpers
def exists(val):
    return val is not None


def eval_decorator(fn):
    def inner(model, *args, **kwargs):
        was_training = model.training
        model.eval()
        out = fn(model, *args, **kwargs)
        model.train(was_training)
        return out

    return inner


# top k filtering


def top_k(logits, thres=0.9):
    k = int((1 - thres) * logits.shape[-1])
    val, ind = torch.topk(logits, k)
    probs = torch.full_like(logits, float("-inf"))
    probs.scatter_(1, ind, val)
    return probs


# normalization
# they use layernorm without bias, something that pytorch does not offer


class LayerNorm(nn.Module):
    def __init__(self, dim):
        super().__init__()
        self.gamma = nn.Parameter(torch.ones(dim))
        self.register_buffer("beta", torch.zeros(dim))

    def forward(self, x):
        return F.layer_norm(x, x.shape[-1:], self.gamma, self.beta)


# residual
# normalization
class RMSNorm(nn.Module):
    def __init__(self, dim, eps=1e-8):
        super().__init__()
        self.scale = dim**-0.5
        self.eps = eps
        self.g = nn.Parameter(torch.ones(dim))

    def forward(self, x):
        norm = torch.norm(x, dim=-1, keepdim=True) * self.scale
        return x / norm.clamp(min=self.eps) * self.g


class Residual(nn.Module):
    def __init__(self, fn):
        super().__init__()
        self.fn = fn

    def forward(self, x):
        return self.fn(x) + x


# rotary positional embedding
# https://arxiv.org/abs/2104.09864


class RotaryEmbedding(nn.Module):
    def __init__(self, dim):
        super().__init__()
        inv_freq = 1.0 / (10000 ** (torch.arange(0, dim, 2).float() / dim))
        self.register_buffer("inv_freq", inv_freq)

    def forward(self, max_seq_len, *, device):
        seq = torch.arange(
            max_seq_len, device=device, dtype=self.inv_freq.dtype
        )
        freqs = einsum("i , j -> i j", seq, self.inv_freq)
        return torch.cat((freqs, freqs), dim=-1)


def rotate_half(x):
    x = rearrange(x, "... (j d) -> ... j d", j=2)
    x1, x2 = x.unbind(dim=-2)
    return torch.cat((-x2, x1), dim=-1)


def apply_rotary_pos_emb(pos, t):
    return (t * pos.cos()) + (rotate_half(t) * pos.sin())


# classic Noam Shazeer paper, except here they use SwiGLU instead of the more popular GEGLU for gating the feedforward
# https://arxiv.org/abs/2002.05202


class SwiGLU(nn.Module):
    def forward(self, x):
        x, gate = x.chunk(2, dim=-1)
        return F.silu(gate) * x


# parallel attention and feedforward with residual
# discovered by Wang et al + EleutherAI from GPT-J fame

# Assuming necessary imports like RotaryEmbedding, SwiGLU, etc. are present


def FeedForward(dim, hidden_dim, dropout=0.0):
    return nn.Sequential(
        nn.LayerNorm(dim),
        nn.Linear(dim, hidden_dim),
        nn.GELU(),
        nn.Linear(hidden_dim, dim),
        nn.Dropout(dropout),
    )


class ParallelTransformerBlock(nn.Module):
    def __init__(self, dim, dim_head=64, heads=8, ff_mult=4):
        super().__init__()
        self.norm = LayerNorm(dim)

        attn_inner_dim = dim_head * heads
        ff_inner_dim = dim * ff_mult
        self.fused_dims = (
            attn_inner_dim,
            dim_head,
            dim_head,
            (ff_inner_dim * 2),
        )

        self.heads = heads
        self.scale = dim_head**-0.5
        self.rotary_emb = RotaryEmbedding(dim_head)

        self.fused_attn_ff_proj = nn.Linear(
            dim, sum(self.fused_dims), bias=False
        )
        self.attn_out = nn.Linear(attn_inner_dim, dim, bias=False)

        self.ff_out = nn.Sequential(
            SwiGLU(), nn.Linear(ff_inner_dim, dim, bias=False)
        )

        # for caching causal mask and rotary embeddings

        self.register_buffer("mask", None, persistent=False)
        self.register_buffer("pos_emb", None, persistent=False)

    def get_mask(self, n, device):
        if self.mask is not None and self.mask.shape[-1] >= n:
            return self.mask[:n, :n]

        mask = torch.ones((n, n), device=device, dtype=torch.bool).triu(1)
        self.register_buffer("mask", mask, persistent=False)
        return mask

    def get_rotary_embedding(self, n, device):
        if self.pos_emb is not None and self.pos_emb.shape[-2] >= n:
            return self.pos_emb[:n]

        pos_emb = self.rotary_emb(n, device=device)
        self.register_buffer("pos_emb", pos_emb, persistent=False)
        return pos_emb

    def forward(self, x):
        """
        einstein notation
        b - batch
        h - heads
        n, i, j - sequence length (base sequence length, source, target)
        d - feature dimension
        """

        n, device, h = x.shape[1], x.device, self.heads

        # pre layernorm

        x = self.norm(x)

        # attention queries, keys, values, and feedforward inner

        q, k, v, ff = self.fused_attn_ff_proj(x).split(self.fused_dims, dim=-1)

        # split heads
        # they use multi-query single-key-value attention, yet another Noam Shazeer paper
        # they found no performance loss past a certain scale, and more efficient decoding obviously
        # https://arxiv.org/abs/1911.02150

        q = rearrange(q, "b n (h d) -> b h n d", h=h)

        # rotary embeddings

        positions = self.get_rotary_embedding(n, device)
        q, k = map(lambda t: apply_rotary_pos_emb(positions, t), (q, k))

        # scale

        q = q * self.scale

        # similarity

        sim = einsum("b h i d, b j d -> b h i j", q, k)

        # causal mask

        causal_mask = self.get_mask(n, device)
        sim = sim.masked_fill(causal_mask, -torch.finfo(sim.dtype).max)

        # attention

        attn = sim.softmax(dim=-1)

        # aggregate values

        out = einsum("b h i j, b j d -> b h i d", attn, v)

        # merge heads

        out = rearrange(out, "b h n d -> b n (h d)")
        return self.attn_out(out) + self.ff_out(ff)


# Transformer
class Transformer(nn.Module):
    def __init__(
        self,
        dim,
        depth,
        heads,
        dim_head,
        ff_mult=4,
    ):
        super().__init__()
        self.layers = nn.ModuleList([])

        self.feedforward = (FeedForward(dim, dim, dropout=0.1),)

        for _ in range(depth):
            self.layers.append(
                nn.ModuleList(
                    [
                        ParallelTransformerBlock(dim, dim_head, heads, ff_mult),
                        FeedForward(dim, dim, dropout=0.1),
                    ]
                )
            )

    def forward(self, x):
        for block, ff in self.layers:
            x = block(x) + x
            x = ff(x) + x
        return x


# classes


class SimpleTransformer(nn.Module):
    """

    Simple Transformer

    Args:
        dim (int): The feature dimension.
        depth (int): The depth of the transformer.
        num_tokens (int): The number of tokens.
        dim_head (int): The dimension of the heads.
        heads (int): The number of heads.
        ff_mult (int): The feedforward multiplier.


    Attributes:
        emb (nn.Embedding): The embedding layer.
        transformer (Transformer): The transformer.
        to_logits (nn.Sequential): The logits layer.

    Example:
        >>> module = SimpleTransformer(512, 6, 20000)
        >>> x = torch.randn(2, 1024).long()
        >>> y = module(x)
        >>> y.shape
        torch.Size([2, 1024, 20000])


    """

    def __init__(
        self,
        dim,
        depth,
        num_tokens,
        dim_head=64,
        heads=8,
        ff_mult=4,
    ):
        super().__init__()
        self.emb = nn.Embedding(num_tokens, dim)

        self.transformer = Transformer(dim, depth, heads, dim_head, ff_mult)

        self.to_logits = nn.Sequential(RMSNorm(dim), nn.Linear(dim, num_tokens))

    def forward(self, x):
        """Forward method implementation."""
        x = self.emb(x)
        x = self.transformer(x)
        return self.to_logits(x)


# autoregressive wrapper for generation
class AutoregressiveWrapper(nn.Module):
    """
    Autoregressive Wrapper

    Args:
        net (nn.Module): The module.
        max_seq_len (int): The maximum sequence length.
        pad_value (int): The pad value.

    Example:
        >>> module = AutoregressiveWrapper(nn.Linear(10, 10))
        >>> x = torch.randn(2, 1024).long()
        >>> y = module(x)
        >>> y.shape
        torch.Size([2, 1024, 10])

    """

    def __init__(self, net, max_seq_len=2048, pad_value=0):
        super().__init__()
        self.max_seq_len = max_seq_len
        self.pad_value = pad_value
        self.net = net

    @torch.no_grad()
    @eval_decorator
    def generate(
        self,
        start_tokens,
        seq_len,
        eos_token=None,
        temperature=1.0,
        filter_thres=0.9,
        **kwargs,
    ):
        """
        Args:

            start_tokens (torch.Tensor): The start tokens.
            seq_len (int): The sequence length.
            eos_token (int): The end of sequence token.
            temperature (float): The temperature.
            filter_thres (float): The filter threshold.
            kwargs (dict): The keyword arguments.

        Returns:
            torch.Tensor: The generated tokens.

        Example:
            >>> module = AutoregressiveWrapper(nn.Linear(10, 10))
            >>> x = torch.randn(2, 1024).long()
            >>> y = module(x)
            >>> y.shape
            torch.Size([2, 1024, 10])


        """
        # einops conflicts with ruff, so noqa on next line
        b, t, device = *start_tokens.shape, start_tokens.device  # noqa F841

        out = start_tokens

        for _ in range(seq_len):
            logits = self.net(out, **kwargs)[:, -1, :]

            filtered_logits = top_k(logits, thres=filter_thres)
            probs = F.softmax(filtered_logits / temperature, dim=-1)

            sample = torch.multinomial(probs, 1)

            out = torch.cat((out, sample), dim=-1)

            if exists(eos_token):
                is_eos_token = out == eos_token

                if is_eos_token.any(dim=-1).all():
                    # mask out everything after the eos tokens
                    shifted_is_eos_tokens = F.pad(is_eos_token, (1, -1))
                    mask = shifted_is_eos_tokens.float().cumsum(dim=-1) >= 1
                    out = out.masked_fill(mask, self.pad_value)
                    break

        out = out[:, t:]
        return out

    def forward(self, x, **kwargs):
        """Forward method implementation."""
        x_inp, x_labels = x[:, :-1], x[:, 1:]
        logits = self.net(x_inp, **kwargs)
        return F.cross_entropy(rearrange(logits, "b c n -> b n c"), x_labels)



# Copyright (c) 2022 Agora
# Licensed under The MIT License [see LICENSE for details]

import torch.nn as nn
from timm.models.layers import drop_path


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks)."""

    def __init__(self, drop_prob=None):
        super().__init__()
        self.drop_prob = drop_prob

    def forward(self, x):
        return drop_path(x, self.drop_prob, self.training)

    def extra_repr(self):
        return f"p={self.drop_prob}"



from einops.layers.torch import EinMix as Mix
from torch import nn


class VisionWeightedPermuteMLP(nn.Module):
    """
    VisionWeightedPermuteMLP module applies weighted permutation to the input tensor
    based on its spatial dimensions (height and width) and channel dimension.

    Args:
        H (int): Height of the input tensor.
        W (int): Width of the input tensor.
        C (int): Number of channels in the input tensor.
        seg_len (int): Length of each segment to divide the channels into.

    Attributes:
        mlp_c (Mix): MLP module for channel dimension permutation.
        mlp_h (Mix): MLP module for height dimension permutation.
        mlp_w (Mix): MLP module for width dimension permutation.
        proj (nn.Linear): Linear projection layer.

    """

    def __init__(self, H, W, C, seg_len):
        super().__init__()
        assert (
            C % seg_len == 0
        ), f"can't divide {C} into segments of length {seg_len}"
        self.mlp_c = Mix(
            "b h w c -> b h w c0",
            weight_shape="c c0",
            bias_shape="c0",
            c=C,
            c0=C,
        )
        self.mlp_h = Mix(
            "b h w (n c) -> b h0 w (n c0)",
            weight_shape="h c h0 c0",
            bias_shape="h0 c0",
            h=H,
            h0=H,
            c=seg_len,
            c0=seg_len,
        )
        self.mlp_w = Mix(
            "b h w (n c) -> b h w0 (n c0)",
            weight_shape="w c w0 c0",
            bias_shape="w0 c0",
            w=W,
            w0=W,
            c=seg_len,
            c0=seg_len,
        )
        self.proj = nn.Linear(C, C)

    def forward(self, x):
        """
        Forward pass of the VisionWeightedPermuteMLP module.

        Args:
            x (torch.Tensor): Input tensor of shape (batch_size, C, H, W).

        Returns:
            torch.Tensor: Output tensor of shape (batch_size, C, H, W).

        """
        x = self.mlp_c(x) + self.mlp_h(x) + self.mlp_w(x)
        return self.proj(x)



import torch.nn.functional as F
from torch import nn


class SimpleRMSNorm(nn.Module):
    """
    SimpleRMSNorm

    Args:
        dim (int): dimension of the embedding

    Usage:
    We can use SimpleRMSNorm as a layer in a neural network as follows:
        >>> x = torch.randn(1, 10, 512)
        >>> simple_rms_norm = SimpleRMSNorm(dim=512)
        >>> simple_rms_norm(x).shape
        torch.Size([1, 10, 512])

    """

    def __init__(self, dim):
        super().__init__()
        self.scale = dim**-0.5

    def forward(self, x):
        """Forward method of SimpleRMSNorm"""
        return F.normalize(x, dim=-1) * self.scale



import copy
from pathlib import Path

import torch
import torch.nn.functional as F
from beartype import beartype
from einops import rearrange, repeat
from einops.layers.torch import Rearrange
from torch import nn


def log(t, eps=1e-10):
    return torch.log(t.clamp(min=eps))


def exists(val):
    return val is not None


def gumbel_noise(t):
    noise = torch.zeros_like(t).uniform_(0, 1)
    return -log(-log(noise))


def gumbel_sample(t, temperature=1.0, dim=-1):
    return ((t / max(temperature, 1e-10)) + gumbel_noise(t)).argmax(dim=dim)


def masked_mean(seq, mask=None, dim=1, keepdim=False):
    if not exists(mask):
        return seq.mean(dim=dim)

    if seq.ndim == 3:
        mask = rearrange(mask, "b n -> b n 1")

    masked_seq = seq.masked_fill(~mask, 0.0)
    numer = masked_seq.sum(dim=dim, keepdim=keepdim)
    denom = mask.sum(dim=dim, keepdim=keepdim)

    masked_mean = numer / denom.clamp(min=1e-3)
    masked_mean = masked_mean.masked_fill(denom == 0, 0.0)
    return masked_mean


@beartype
class RewardModel(nn.Module):
    """
    Reward Model:

    Parameters
    ----------
    model : nn.Module
        Model to use
    dropout : float
        Dropout
    num_binned_output : int
        Number of binned output
    use_lora : bool
        Whether to use lora
    lora_r : int
        Lora r
    reward_lora_scope : str
        Reward lora scope

    Returns
    -------
    pred : torch.Tensor

    Usage:
    ------
    >>> model = RewardModel(model, num_binned_output=10)
    >>> pred = model(x)


    """

    def __init__(
        self,
        model,
        dropout=0.1,
        num_binned_output=0.0,
        use_lora=True,
        lora_r=8,
        reward_lora_scope="reward",
    ):
        super().__init__()

        self.model = copy.deepcopy(model)
        self.model.set_dropout(dropout)

        self.reward_lora_scope = reward_lora_scope if use_lora else None

        if exists(self.reward_lora_scope):
            self.model.add_finetune_params(reward_lora_scope, lora_r=lora_r)

        dim = model.dim

        self.binned_output = num_binned_output > 1

        self.prompt_embed = nn.Parameter(torch.zeros(1, 1, dim))
        self.response_embed = nn.Parameter(torch.zeros(1, 1, dim))

        if self.binned_output:
            self.to_pred = nn.Linear(dim, num_binned_output)

        else:
            self.to_pred = nn.Sequential(
                nn.Linear(dim, 1, bias=False), Rearrange("... 1 -> ...")
            )

    def load(self, path):
        """Load model"""
        path = Path(path)
        assert path.exists()
        self.load_state_dict(torch.load(path, weights_only=True))

    def finetune_parameters(self):
        """Finetune parameters"""
        return [
            *self.to_pred.parameters(),
            *(
                self.model.finetune_parameters(self.reward_lora_scope)
                if exists(self.reward_lora_scope)
                else self.model.parameters()
            ),
        ]

    def forward(
        self,
        x,
        mask=None,
        prompt_mask=None,
        prompt_lengths=None,
        labels=None,
        sample=None,
        sample_temperature=1.0,
        disable_lora=False,
    ):
        assert not (exists(prompt_mask) and exists(prompt_lengths))

        if exists(prompt_lengths):
            batch, seq_len = x.shape
            arange = torch.arange(seq_len, device=x.device)
            prompt_mask = repeat(arange, "n -> b n", b=batch) < rearrange(
                prompt_lengths, "b -> b 1"
            )

        # model need to know what is prompt and what is response

        extra_embed = None

        if exists(prompt_mask):
            extra_embed = torch.where(
                rearrange(prompt_mask, "b n -> b n 1"),
                self.prompt_embed,
                self.response_embed,
            )

        embeds = self.model(
            x,
            extra_embed=extra_embed,
            return_only_embedding=True,
            disable_lora=disable_lora,
            finetune_scope=self.reward_lora_scope,
        )

        pooled = masked_mean(embeds, mask, dim=1)
        pred = self.to_pred(pooled)

        if sample and self.binned_output:
            assert not exists(labels)
            pred = gumbel_sample(pred, temperature=sample_temperature, dim=-1)

        if not exists(labels):
            return pred

        if not self.binned_output:
            return F.mse_loss(pred, labels)

        if not self.binned_output:
            return F.mse_loss(pred, labels)

        return F.cross_entropy(pred, labels)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zeta',
 'zeta.cli',
 'zeta.cloud',
 'zeta.models',
 'zeta.nn',
 'zeta.nn.attention',
 'zeta.nn.biases',
 'zeta.nn.embeddings',
 'zeta.nn.masks',
 'zeta.nn.modules',
 'zeta.nn.modules.xmoe',
 'zeta.ops',
 'zeta.optim',
 'zeta.quant',
 'zeta.rl',
 'zeta.structs',
 'zeta.tokenizers',
 'zeta.training',
 'zeta.utils']

package_data = \
{'': ['*']}

install_requires = \
['accelerate==0.27.2',
 'argparse>=1.4.0,<2.0.0',
 'beartype==0.17.2',
 'bitsandbytes==0.42.0',
 'colt5-attention==0.10.19',
 'datasets',
 'einops-exts==0.0.4',
 'einops==0.7.0',
 'fairseq==0.12.2',
 'jax',
 'jaxlib',
 'lion-pytorch==0.1.2',
 'loguru',
 'numexpr',
 'pytest==8.0.1',
 'rich==13.7.0',
 'scipy==1.9.3',
 'sentencepiece==0.1.99',
 'skypilot==0.4.1',
 'tensorflow',
 'tiktoken==0.6.0',
 'timm==0.9.16',
 'tokenmonster==1.1.12',
 'torch==2.2.0',
 'torchdiffeq==0.2.3',
 'torchfix',
 'torchvision==0.17.0',
 'tqdm==4.66.2',
 'transformers==4.36.2',
 'typing==3.7.4.3',
 'vector-quantize-pytorch==1.14.1']

entry_points = \
{'console_scripts': ['zeta = zeta.cli.main:main']}

setup_kwargs = {
    'name': 'zetascale',
    'version': '2.2.1',
    'description': 'Transformers at zeta scales',
    'long_description': '[![Multi-Modality](images/agorabanner.png)](https://discord.gg/qUtxnK2NMf)\n\n![Zeta banner](images/zeta.png)\nBuild SOTA AI Models 80% faster with modular, high-performance, and scalable building blocks!\n\n[![Docs](https://readthedocs.org/projects/zeta/badge/)](https://zeta.readthedocs.io)\n\n<p>\n  <a href="https://github.com/kyegomez/zeta/blob/main/LICENSE"><img alt="MIT License" src="https://img.shields.io/badge/license-MIT-blue.svg" /></a>\n  <a href="https://pypi.org/project/zetascale"><img alt="MIT License" src="https://badge.fury.io/py/zetascale.svg" /></a>\n</p>\n\n[![GitHub issues](https://img.shields.io/github/issues/kyegomez/zeta)](https://github.com/kyegomez/zeta/issues) [![GitHub forks](https://img.shields.io/github/forks/kyegomez/zeta)](https://github.com/kyegomez/zeta/network) [![GitHub stars](https://img.shields.io/github/stars/kyegomez/zeta)](https://github.com/kyegomez/zeta/stargazers) [![GitHub license](https://img.shields.io/github/license/kyegomez/zeta)](https://github.com/kyegomez/zeta/blob/main/LICENSE)[![GitHub star chart](https://img.shields.io/github/stars/kyegomez/zeta?style=social)](https://star-history.com/#kyegomez/zeta)[![Dependency Status](https://img.shields.io/librariesio/github/kyegomez/zeta)](https://libraries.io/github/kyegomez/zeta) [![Downloads](https://static.pepy.tech/badge/zeta/month)](https://pepy.tech/project/zeta)\n\n[![Join the Agora discord](https://img.shields.io/discord/1110910277110743103?label=Discord&logo=discord&logoColor=white&style=plastic&color=d7b023)![Share on Twitter](https://img.shields.io/twitter/url/https/twitter.com/cloudposse.svg?style=social&label=Share%20%40kyegomez/zeta)](https://twitter.com/intent/tweet?text=Check%20out%20this%20amazing%20AI%20project:%20&url=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fzeta) [![Share on Facebook](https://img.shields.io/badge/Share-%20facebook-blue)](https://www.facebook.com/sharer/sharer.php?u=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fzeta) [![Share on LinkedIn](https://img.shields.io/badge/Share-%20linkedin-blue)](https://www.linkedin.com/shareArticle?mini=true&url=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fzeta&title=&summary=&source=)\n\n[![Share on Reddit](https://img.shields.io/badge/-Share%20on%20Reddit-orange)](https://www.reddit.com/submit?url=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fzeta&title=zeta%20-%20the%20future%20of%20AI) [![Share on Hacker News](https://img.shields.io/badge/-Share%20on%20Hacker%20News-orange)](https://news.ycombinator.com/submitlink?u=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fzeta&t=zeta%20-%20the%20future%20of%20AI) [![Share on Pinterest](https://img.shields.io/badge/-Share%20on%20Pinterest-red)](https://pinterest.com/pin/create/button/?url=https%3A%2F%2Fgithub.com%2Fkyegomez%2Fzeta&media=https%3A%2F%2Fexample.com%2Fimage.jpg&description=zeta%20-%20the%20future%20of%20AI) [![Share on WhatsApp](https://img.shields.io/badge/-Share%20on%20WhatsApp-green)](https://api.whatsapp.com/send?text=Check%20out%20zeta%20-%20the%20future%20of%20AI%20%23zeta%20%23AI%0A%0Ahttps%3A%2F%2Fgithub.com%2Fkyegomez%2Fzeta)\n\nAfter building out thousands of neural nets and facing the same annoying bottlenecks of chaotic codebases with no modularity and low performance modules, Zeta needed to be born to enable me and others to quickly prototype, train, and optimize the latest SOTA neural nets and deploy them into production. \n\nZeta places a radical emphasis on useability, modularity, and performance. Zeta is now currently employed in 100s of models across my github and across others. \nGet started below and LMK if you want my help building any model, I\'m here for you 😊 💜 \n\n\n# Install\n\n`$ pip3 install -U zetascale`\n\n# Usage\n\n## Starting Your Journey\n\nCreating a model empowered with the aforementioned breakthrough research features is a breeze. Here\'s how to quickly materialize the renowned Flash Attention\n\n```python\nimport torch\n\nfrom zeta.nn import FlashAttention\n\nq = torch.randn(2, 4, 6, 8)\nk = torch.randn(2, 4, 10, 8)\nv = torch.randn(2, 4, 10, 8)\n\nattention = FlashAttention(causal=False, dropout=0.1, flash=True)\noutput = attention(q, k, v)\n\nprint(output.shape)\n```\n\n\n\n### `SwiGLU`\n- Powers Transformer models\n```python\nimport torch\n\nfrom zeta.nn import SwiGLUStacked\n\nx = torch.randn(5, 10)\nswiglu = SwiGLUStacked(10, 20)\nswiglu(x).shape\n```\n\n### ```RelativePositionBias```\n- ```RelativePositionBias``` quantizes the distance between two positions into a certain number of buckets and then uses an embedding to get the relative position bias. This mechanism aids in the attention mechanism by providing biases based on relative positions between the query and key, rather than relying solely on their absolute positions.\n```python\nimport torch\nfrom torch import nn\n\nfrom zeta.nn import RelativePositionBias\n\n# Initialize the RelativePositionBias module\nrel_pos_bias = RelativePositionBias()\n\n# Example 1: Compute bias for a single batch\nbias_matrix = rel_pos_bias(1, 10, 10)\n\n\n# Example 2: Utilize in conjunction with an attention mechanism\n# NOTE: This is a mock example, and may not represent an actual attention mechanism\'s complete implementation.\nclass MockAttention(nn.Module):\n    def __init__(self):\n        super().__init__()\n        self.rel_pos_bias = RelativePositionBias()\n\n    def forward(self, queries, keys):\n        bias = self.rel_pos_bias(queries.size(0), queries.size(1), keys.size(1))\n        # Further computations with bias in the attention mechanism...\n        return None  # Placeholder\n\n\n# Example 3: Modify default configurations\ncustom_rel_pos_bias = RelativePositionBias(\n    bidirectional=False, num_buckets=64, max_distance=256, num_heads=8\n)\n```\n\n### `FeedForward`\nThe FeedForward module performs a feedforward operation on the input tensor x. It consists of a multi-layer perceptron (MLP) with an optional activation function and LayerNorm. \nUsed in most language, multi-modal, and modern neural networks.\n\n```python\nimport torch\n\nfrom zeta.nn import FeedForward\n\nmodel = FeedForward(256, 512, glu=True, post_act_ln=True, dropout=0.2)\n\nx = torch.randn(1, 256)\n\noutput = model(x)\nprint(output.shape)\n```\n\n### `BitLinear`\n- The BitLinear module performs linear transformation on the input data, followed by quantization and dequantization. The quantization process is performed using the absmax_quantize function, which quantizes the input tensor based on the absolute maximum value, [from the paper](https://arxiv.org/abs/2310.11453)\n```python\nimport torch\nfrom torch import nn\n\nimport zeta.quant as qt\n\n\nclass MyModel(nn.Module):\n    def __init__(self):\n        super().__init__()\n        self.linear = qt.BitLinear(10, 20)\n\n    def forward(self, x):\n        return self.linear(x)\n\n\n# Initialize the model\nmodel = MyModel()\n\n# Create a random tensor of size (128, 10)\ninput = torch.randn(128, 10)\n\n# Perform the forward pass\noutput = model(input)\n\n# Print the size of the output\nprint(output.size())  # torch.Size([128, 20])\n```\n\n### `PalmE`\n- This is an implementation of the multi-modal Palm-E model using a decoder llm as the backbone with an VIT image encoder to process vision, it\'s very similiar to GPT4, Kosmos, RTX2, and many other multi-modality model architectures\n\n```python\nimport torch\n\nfrom zeta.structs import (\n    AutoregressiveWrapper,\n    Decoder,\n    Encoder,\n    Transformer,\n    ViTransformerWrapper,\n)\n\n\nclass PalmE(torch.nn.Module):\n    """\n        PalmE is a transformer architecture that uses a ViT encoder and a transformer decoder.\n\n        Args:\n\n            image_size (int): Size of the image.\n            patch_size (int): Size of the patch.\n            encoder_dim (int): Dimension of the encoder.\n            encoder_depth (int): Depth of the encoder.\n            encoder_heads (int): Number of heads in the encoder.\n            num_tokens (int): Number of tokens.\n            max_seq_len (int): Maximum sequence length.\n            decoder_dim (int): Dimension of the decoder.\n            decoder_depth (int): Depth of the decoder.\n            decoder_heads (int): Number of heads in the decoder.\n            alibi_num_heads (int): Number of heads in the alibi attention.\n            attn_kv_heads (int): Number of heads in the attention key-value projection.\n            use_abs_pos_emb (bool): Whether to use absolute positional embeddings.\n            cross_attend (bool): Whether to cross attend in the decoder.\n            alibi_pos_bias (bool): Whether to use positional bias in the alibi attention.\n            rotary_xpos (bool): Whether to use rotary positional embeddings.\n            attn_flash (bool): Whether to use attention flash.\n            qk_norm (bool): Whether to normalize the query and key in the attention layer.\n\n        Returns:\n\n                torch.Tensor: The output of the model.\n\n        Usage:\n\n    img = torch.randn(1, 3, 256, 256)\n    text = torch.randint(0, 20000, (1, 1024))\n    model = PalmE()\n    output = model(img, text)\n    print(output)\n\n    """\n\n    def __init__(\n        self,\n        image_size=256,\n        patch_size=32,\n        encoder_dim=512,\n        encoder_depth=6,\n        encoder_heads=8,\n        num_tokens=20000,\n        max_seq_len=1024,\n        decoder_dim=512,\n        decoder_depth=6,\n        decoder_heads=8,\n        alibi_num_heads=4,\n        attn_kv_heads=2,\n        use_abs_pos_emb=False,\n        cross_attend=True,\n        alibi_pos_bias=True,\n        rotary_xpos=True,\n        attn_flash=True,\n        qk_norm=True,\n    ):\n        super().__init__()\n\n        # vit architecture\n        self.encoder = ViTransformerWrapper(\n            image_size=image_size,\n            patch_size=patch_size,\n            attn_layers=Encoder(\n                dim=encoder_dim, depth=encoder_depth, heads=encoder_heads\n            ),\n        )\n\n        # palm model architecture\n        self.decoder = Transformer(\n            num_tokens=num_tokens,\n            max_seq_len=max_seq_len,\n            use_abs_pos_emb=use_abs_pos_emb,\n            attn_layers=Decoder(\n                dim=decoder_dim,\n                depth=decoder_depth,\n                heads=decoder_heads,\n                cross_attend=cross_attend,\n                alibi_pos_bias=alibi_pos_bias,\n                alibi_num_heads=alibi_num_heads,\n                rotary_xpos=rotary_xpos,\n                attn_kv_heads=attn_kv_heads,\n                attn_flash=attn_flash,\n                qk_norm=qk_norm,\n            ),\n        )\n\n        # autoregressive wrapper to enable generation of tokens\n        self.decoder = AutoregressiveWrapper(self.decoder)\n\n    def forward(self, img: torch.Tensor, text: torch.Tensor):\n        """Forward pass of the model."""\n        try:\n            encoded = self.encoder(img, return_embeddings=True)\n            return self.decoder(text, context=encoded)\n        except Exception as error:\n            print(f"Failed in forward method: {error}")\n            raise\n\n\n# Usage with random inputs\nimg = torch.randn(1, 3, 256, 256)\ntext = torch.randint(0, 20000, (1, 1024))\n\n# Initiliaze the model\nmodel = PalmE()\noutput = model(img, text)\nprint(output)\n```\n\n\n### `Unet`\nUnet is a famous convolutional neural network architecture originally used for biomedical image segmentation but soon became the backbone of the generative AI Mega-revolution. The architecture comprises two primary pathways: downsampling and upsampling, followed by an output convolution. Due to its U-shape, the architecture is named U-Net. Its symmetric architecture ensures that the context (from downsampling) and the localization (from upsampling) are captured effectively.\n\n```python\nimport torch\n\nfrom zeta.nn import Unet\n\n# Initialize the U-Net model\nmodel = Unet(n_channels=1, n_classes=2)\n\n# Random input tensor with dimensions [batch_size, channels, height, width]\nx = torch.randn(1, 1, 572, 572)\n\n# Forward pass through the model\ny = model(x)\n\n# Output\nprint(f"Input shape: {x.shape}")\nprint(f"Output shape: {y.shape}")\n```\n\n\n### `VisionEmbeddings`\nThe VisionEmbedding class is designed for converting images into patch embeddings, making them suitable for processing by transformer-based models. This class plays a crucial role in various computer vision tasks and enables the integration of vision data into transformer architectures!\n\n```python\nimport torch\n\nfrom zeta.nn import VisionEmbedding\n\n# Create an instance of VisionEmbedding\nvision_embedding = VisionEmbedding(\n    img_size=224,\n    patch_size=16,\n    in_chans=3,\n    embed_dim=768,\n    contain_mask_token=True,\n    prepend_cls_token=True,\n)\n\n# Load an example image (3 channels, 224x224)\ninput_image = torch.rand(1, 3, 224, 224)\n\n# Perform image-to-patch embedding\noutput = vision_embedding(input_image)\n\n# The output now contains patch embeddings, ready for input to a transformer model\n```\n\n\n### `niva`\n- Niva focuses on weights of certain layers (specified by quantize_layers). Ideal for models where runtime activation is variable. 👁️ Example Layers: nn.Embedding, nn.LSTM. \n\n```python\nimport torch\n\nfrom zeta import niva\n\n# Load a pre-trained model\nmodel = YourModelClass()\n\n# Quantize the model dynamically, specifying layers to quantize\nniva(\n    model=model,\n    model_path="path_to_pretrained_model_weights.pt",\n    output_path="quantized_model.pt",\n    quant_type="dynamic",\n    quantize_layers=[nn.Linear, nn.Conv2d],\n    dtype=torch.qint8,\n)\n```\n\n\n### `FusedDenseGELUDense`\n- Increase model speed by 2x with this module that fuses together 2 hyper-optimized dense ops from bits and bytes and a gelu together!\n\n```python\nimport torch\n\nfrom zeta.nn import FusedDenseGELUDense\n\nx = torch.randn(1, 512)\nmodel = FusedDenseGELUDense(512, 1024)\nout = model(x)\nout.shape\n```\n\n\n### `FusedDropoutLayerNorm`\n- FusedDropoutLayerNorm is a fused kernel of dropout and layernorm to speed up FFNs or MLPS by 2X\n\n```python\nimport torch\nfrom torch import nn\n\nfrom zeta.nn import FusedDropoutLayerNorm\n\n# Initialize the module\nmodel = FusedDropoutLayerNorm(dim=512)\n\n# Create a sample input tensor\nx = torch.randn(1, 512)\n\n# Forward pass\noutput = model(x)\n\n# Check output shape\nprint(output.shape)  # Expected: torch.Size([1, 512])\n```\n\n\n### `Mamba`\n- Pytorch implementation of the new SSM model architecture Mamba\n\n```python\nimport torch\n\nfrom zeta.nn import MambaBlock\n\n# Initialize Mamba\nblock = MambaBlock(dim=64, depth=1)\n\n# Random input\nx = torch.randn(1, 10, 64)\n\n# Apply the model to the block\ny = block(x)\n\nprint(y.shape)\n# torch.Size([1, 10, 64])\n```\n\n### `FiLM`\n\n```python\nimport torch\n\nfrom zeta.nn import Film\n\n# Initialize the Film layer\nfilm_layer = Film(dim=128, hidden_dim=64, expanse_ratio=4)\n\n# Create some dummy data for conditions and hiddens\nconditions = torch.randn(10, 128)  # Batch size is 10, feature size is 128\nhiddens = torch.randn(\n    10, 1, 128\n)  # Batch size is 10, sequence length is 1, feature size is 128\n\n# Pass the data through the Film layer\nmodulated_features = film_layer(conditions, hiddens)\n\n# Print the shape of the output\nprint(modulated_features.shape)  # Should be [10, 1, 128]\n```\n\n### `hyper_optimize`\n- A single wrapper for torch.fx, torch.script, torch.compile, dynamic quantization, mixed precision through torch.amp, with execution time metrics all in once place!\n```python\nimport torch\n\nfrom zeta.nn import hyper_optimize\n\n\n@hyper_optimize(\n    torch_fx=False,\n    torch_script=False,\n    torch_compile=True,\n    quantize=True,\n    mixed_precision=True,\n    enable_metrics=True,\n)\ndef model(x):\n    return x @ x\n\n\nout = model(torch.randn(1, 3, 32, 32))\nprint(out)\n```\n\n\n### DPO - Direct Policy Optimization\nDirect Policy Optimization employed for many RLHF applications for LLMs.\n\n```python\nimport torch\nfrom torch import nn\n\nfrom zeta.rl import DPO\n\n\n# Define a simple policy model\nclass PolicyModel(nn.Module):\n    def __init__(self, input_dim, output_dim):\n        super().__init__()\n        self.fc = nn.Linear(input_dim, output_dim)\n\n    def forward(self, x):\n        return self.fc(x)\n\n\ninput_dim = 10\noutput_dim = 5\npolicy_model = PolicyModel(input_dim, output_dim)\n\n# Initialize DPO with the policy model\ndpo_model = DPO(model=policy_model, beta=0.1)\n\n# Sample preferred and unpreferred sequences\npreferred_seq = torch.randint(0, output_dim, (3, input_dim))\nunpreferred_seq = torch.randint(0, output_dim, (3, input_dim))\n\n# Compute loss\nloss = dpo_model(preferred_seq, unpreferred_seq)\nprint(loss)\n```\n\n\n### ZetaCloud\nTrain or finetune any model on any cluster in 1 click with zetacloud, just pass in your file and the GPU type and quantity you want! To gain access first `pip install zetascale` then run `zeta -h` in the terminal. [Here is the docs for more](https://zeta.apac.ai/en/latest/zeta/cloud/main/)\n\n- Flexible Pricing with pooling from many clouds\n- Easy Deployment with 1 click\n- Various options for cloud providers!\n\n```bash\nZetacloud CLI\n\noptions:\n  -h, --help            show this help message and exit\n  -t TASK_NAME, --task_name TASK_NAME\n                        Task name\n  -c CLUSTER_NAME, --cluster_name CLUSTER_NAME\n                        Cluster name\n  -cl CLOUD, --cloud CLOUD\n                        Cloud provider\n  -g GPUS, --gpus GPUS  GPUs\n  -f FILENAME, --filename FILENAME\n                        Filename\n  -s, --stop            Stop flag\n  -d, --down            Down flag\n  -sr, --status_report  Status report flag\n\n```\n\n- A simple run example code would be like:\n\n```bash\nzeta -f train.py -g A100:8\n```\n----\n\n\n# Documentation\nAll classes must have documentation if you see a class or function without documentation then please report it to me at kye@apac.ai,\n\nDocumentation is at [zeta.apac.ai](https://zeta.apac.ai/)\n\n\n-------\n\n\n# Running tests\nYou should install the pre-commit hooks with pre-commit install. This will run the linter, mypy, and a subset of the tests on every commit.\n\nFor more examples on how to run the full test suite please refer to the CI workflow.\n\nSome examples of running tests locally:\n\n```bash\npython3 -m pip install -e \'.[testing]\'  # install extra deps for testing\npython3 -m pytest tests/                 # whole test suite\n```\n----\n\n## Community\n\nJoin our growing community around the world, for real-time support, ideas, and discussions on how to build better models 😊 \n\n- View our official [Docs](https://zeta.apac.ai)\n- Chat live with us on [Discord](https://discord.gg/kS3rwKs3ZC)\n- Follow us on [Twitter](https://twitter.com/kyegomez)\n- Connect with us on [LinkedIn](https://www.linkedin.com/company/the-swarm-corporation)\n- Visit us on [YouTube](https://www.youtube.com/channel/UC9yXyitkbU_WSy7bd_41SqQ)\n- [Join the Swarms community on Discord!](https://discord.gg/AJazBmhKnr)\n\n---\n\n# 🤝 Schedule a 1-on-1 Session\nWant to train a custom AI model for a real-world task like General Multi-Modal Models, Facial Recognitions, Drug Discovery, Humanoid Robotics? I\'ll help you create the model architecture then train the model and then optimize it to meet your quality assurance standards.\n\nBook a [1-on-1 Session with Kye here.](https://calendly.com/apacai/agora), the Creator, to discuss any issues, provide feedback, or explore how we can improve Zeta for you or help you build your own custom models!\n\n## 🫶 Contributions:\n\nThe easiest way to contribute is to pick any issue with the `good first issue` tag 💪. Read the Contributing guidelines [here](/CONTRIBUTING.md). Bug Report? [File here](https://github.com/kyegomez/zeta/issues/new/choose) | Feature Request? [File here](https://github.com/kyegomez/zeta/issues/new/choose)\n\nZeta is an open-source project, and contributions are VERY welcome. If you want to contribute, you can create new features, fix bugs, or improve the infrastructure. Please refer to the [CONTRIBUTING.md](https://github.com/kyegomez/zeta/blob/master/CONTRIBUTING.md) and our [contributing board](https://github.com/users/kyegomez/projects/1) to participate in Roadmap discussions!\n\n<a href="https://github.com/kyegomez/zeta/graphs/contributors">\n  <img src="https://contrib.rocks/image?repo=kyegomez/zeta" />\n</a>\n\n----\n\n## Accelerate Backlog\nHelp us accelerate our backlog by supporting us financially! Note, we\'re an open source corporation and so all the revenue we generate is through donations at the moment ;)\n\n<a href="https://polar.sh/kyegomez"><img src="https://polar.sh/embed/fund-our-backlog.svg?org=kyegomez" /></a>\n\n\n# License \n- Apache\n',
    'author': 'Zeta Team',
    'author_email': 'kye@apac.ai',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/kyegomez/zeta',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Cloud"""

import logging
from typing import Any

from sky import AWS, Resources

from zeta.cloud.sky_api import SkyInterface

skyapi = SkyInterface(stream_logs_enabled=True)


# Logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


def zetacloud(
    task_name: str = None,
    cluster_name: str = "ZetaTrainingRun",
    setup: str = "pip install -r requirements.txt",
    cloud: Any = AWS(),
    gpus: str = "V100:4",
    filename: str = "train.py",
    stop: bool = False,
    down: bool = False,
    status_report: bool = False,
    *args,
    **kwargs,
):
    """zetacloud

    Args:
        task_name (str, optional): _description_. Defaults to None.
        cluster_name (str, optional): _description_. Defaults to "[ZetaTrainingRun]".
        cloud (Any, optional): _description_. Defaults to AWS().
        gpus (str, optional): _description_. Defaults to None.
    """
    try:
        task = skyapi.create_task(
            name=task_name,
            setup=setup,
            run=f"python {filename}",
            workdir=".",
        )
        logger.info(f"Task: {task} has been created")

        # Set the resources
        task.set_resources(Resources(accelerators=gpus))
        # logger.info(f"Resources: {task.resources} have been set")

        # Execute the task on the cluster
        execution = skyapi.launch(task, cluster_name)
        print(execution)
        logger.info(
            f"Task: {task} has been launched on cluster: {cluster_name}"
        )

        if stop:
            skyapi.stop(cluster_name)
            logger.info(f"Cluster: {cluster_name} has been stopped")

        if down:
            skyapi.down(cluster_name)
            logger.info(f"Cluster: {cluster_name} has been deleted")

        if status_report:
            skyapi.status(cluster_names=[cluster_name])
            logger.info(f"Cluster: {cluster_name} has been reported on")

    except Exception as error:
        print(
            f"There has been an error: {error} the root cause is:"
            f" {error.__cause__}"
        )



from einops.layers.torch import Rearrange, Reduce
from torch import nn


def FastTextNew(vocab_size, embedding_dim, output_dim):
    """
    FastText for language

    Args:
    vocab_size: size of vocabulary
    embedding_dim: dimension of embedding
    output_dim: dimension of output

    Usage:
        net = FastTextNew(
            vocab_size=10,
            embedding_dim=20,
            output_dim=10,
        )
        net(x)

    """
    return nn.Sequential(
        Rearrange("t b -> t b "),
        nn.Embedding(vocab_size, embedding_dim),
        Reduce("t b c -> b c", "mean"),
        nn.Linear(embedding_dim, output_dim),
        Rearrange("b c -> b c"),
    )



from typing import Tuple

import torch
from huggingface_hub import snapshot_download
from PIL import Image
from torch import nn
from torchvision.transforms.v2 import (
    Compose,
    InterpolationMode,
    Normalize,
    Resize,
    ToDtype,
    ToImage,
)


class VisionEncoder(nn.Module):
    """
    Initializes a VisionEncoder object.

    Args:
        size (Tuple, optional): The size of the input image. Defaults to (384, 384).
        model_path (str, optional): The path to the pre-trained vision model. Defaults to "model".
        return_shape (bool, optional): Whether to return the shape of the embedding. Defaults to False.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Examples::
        >>> from zeta.structs import VisionEncoder
        >>> encoder = VisionEncoder()
        >>> embeds = encoder("image.jpg")
        >>> embeds.shape
        torch.Size([1, 512])
    """

    def __init__(
        self,
        size: Tuple = (384, 384),
        model_name: str = "vikhyatk/moondream0",
        return_shape: bool = False,
        *args,
        **kwargs,
    ) -> None:
        super().__init__()
        self.size = size
        self.model_name = model_name
        self.return_shape = return_shape
        model_path = snapshot_download(model_name)

        self.model = torch.jit.load(f"{model_path}/vision.pt").to(
            dtype=torch.float32
        )

        self.preprocess = Compose(
            [
                Resize(size=size, interpolation=InterpolationMode.BICUBIC),
                ToImage(),
                ToDtype(torch.float32, scale=True),
                Normalize(mean=[0.5, 0.5, 0.5], std=[0.5, 0.5, 0.5]),
                *args,
            ]
        )

    def __call__(self, image: Image, *args, **kwargs) -> torch.Tensor:
        """
        Processes an input image and returns its embedding.

        Args:
            image (Image): The input image.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            torch.Tensor: The embedding of the input image.
        """
        image = Image.open(image)
        with torch.no_grad():
            image_vec = self.preprocess(image.convert("RGB")).unsqueeze(0)
            embeds = self.model(image_vec, *args, **kwargs)

            if self.return_shape:
                print(f"Embedding shape: {embeds.shape}")

            return embeds



import math
from functools import partial

import numpy as np
import torch
import torch.nn.functional as F
from einops import rearrange
from torch import nn
from torch.nn import functional as F, init


def blockdiag_butterfly_multiply_reference(x, w1_bfly, w2_bfly, version=2):
    """
    This implementation is slow but more likely to be correct.
    There are 3 implementations, which should all yield the same answer
    Arguments:
        x: (batch, n)
        w1_bfly: (k, q, p), where k = n / p
        w2_bfly: (l, s, r), where l = k * q / r = n * q / (p * r)
    Outputs:
        out: (batch, m), where m = l * s = n * s * q / (p * r)
    """
    if version not in [1, 2, 3]:
        raise NotImplementedError("version must be either 1, 2, or 3")
    batch, n = x.shape
    k, q, p = w1_bfly.shape
    l, s, r = w2_bfly.shape
    assert k * p == n
    assert l * r == k * q

    x_reshaped = rearrange(x, "b (k p) -> b k p", k=k)
    if (
        version == 1
    ):  # Implementation 1 (only works for when k = q = p = l = s = r = sqrt(n))
        assert k == q == p == l == s == r == int(math.sqrt(n))
        return torch.einsum(
            "bkp,kqp,qlk->blq", x_reshaped, w1_bfly, w2_bfly
        ).reshape(batch, n)
    elif version == 2:  # Implementation 2
        out1 = torch.einsum("kqp,bkp->bkq", w1_bfly, x_reshaped)
        out1 = rearrange(
            rearrange(out1, "b k q -> b (k q)"), "b (r l) -> b l r", l=l
        )
        return torch.einsum("lsr,blr->bsl", w2_bfly, out1).reshape(batch, s * l)
    # Implementation 3: most likely to be correct, but it's the slowest
    elif version == 3:
        w1_dense = torch.block_diag(*torch.unbind(w1_bfly, dim=0))
        out1 = F.linear(x, w1_dense)
        out1 = rearrange(out1, "b (r l) -> b (l r)", l=l)
        w2_dense = torch.block_diag(*torch.unbind(w2_bfly, dim=0))
        out2 = F.linear(out1, w2_dense)
        out2 = rearrange(out2, "b (l s) -> b (s l)", l=l)
        return out2


class BlockdiagButterflyMultiply(torch.autograd.Function):
    """This is a faster implementation, with careful memory copies for the fastest
    bmm performance.
    The backward pass is also written manually with careful memory copies.
    Arguments:
        x: (batch, n)
        w1_bfly: (k, q, p), where k = n / p
        w2_bfly: (l, s, r), where l = k * q / r = n * q / (p * r)
    Outputs:
        out: (batch, m), where m = l * s = n * s * q / (p * r)
    """

    @staticmethod
    @torch.cuda.amp.custom_fwd(cast_inputs=torch.bfloat16)
    def forward(ctx, x, w1_bfly, w2_bfly):
        batch_shape, n = x.shape[:-1], x.shape[-1]
        batch_dim = np.prod(batch_shape)
        k, q, p = w1_bfly.shape
        l, s, r = w2_bfly.shape
        assert k * p == n
        assert l * r == k * q
        x_reshaped = x.reshape(batch_dim, k, p).transpose(0, 1)
        out1 = torch.empty(
            batch_dim, k, q, device=x.device, dtype=x.dtype
        ).transpose(0, 1)
        out1 = torch.bmm(x_reshaped, w1_bfly.transpose(-1, -2), out=out1)
        out1 = (
            out1.transpose(0, 1)
            .reshape(batch_dim, r, l)
            .transpose(-1, -2)
            .contiguous()
            .transpose(0, 1)
        )
        out2 = torch.empty(
            batch_dim, l, s, device=x.device, dtype=x.dtype
        ).transpose(0, 1)
        out2 = torch.bmm(out1, w2_bfly.transpose(-1, -2), out=out2)
        out2 = out2.permute(1, 2, 0).reshape(*batch_shape, s * l)
        ctx.save_for_backward(x, w1_bfly, w2_bfly, out1)
        return out2

    @staticmethod
    @torch.cuda.amp.custom_bwd
    def backward(ctx, dout):
        x, w1_bfly, w2_bfly, out1 = ctx.saved_tensors
        batch_shape, n = x.shape[:-1], x.shape[-1]
        batch_dim = np.prod(batch_shape)
        k, q, p = w1_bfly.shape
        l, s, r = w2_bfly.shape
        # assert k * p == n
        # assert l * r == k * q
        dx, dw1_bfly, dw2_bfly = None, None, None
        # dout_reshaped = dout.reshape(batch_dim, sqrtn, sqrtn).permute(2, 1, 0).contiguous()
        dout_reshaped = (
            dout.reshape(batch_dim, s, l).transpose(-1, -2).contiguous()
        )
        dout_reshaped = dout_reshaped.transpose(0, 1)
        if ctx.needs_input_grad[2]:
            # dw2_bfly = torch.empty(l, s, r, device=w2_bfly.device, dtype=w2_bfly.dtype)
            # dw2_bfly = torch.bmm(dout_reshaped.transpose(-1, -2), out1, out=dw2_bfly)
            dw2_bfly = torch.bmm(dout_reshaped.transpose(-1, -2), out1.conj())
        if ctx.needs_input_grad[1] or ctx.needs_input_grad[0]:
            dout1 = torch.empty(
                batch_dim, l, r, device=x.device, dtype=x.dtype
            ).transpose(0, 1)
            dout1 = torch.bmm(dout_reshaped, w2_bfly.conj(), out=dout1)
            dout1 = (
                dout1.transpose(0, 1)
                .transpose(-1, -2)
                .contiguous()
                .reshape(batch_dim, k, q)
                .transpose(0, 1)
            )
            # dout1 = dout1.permute(1, 2, 0).contiguous().transpose(0, 1)
            if ctx.needs_input_grad[0]:
                dx = torch.empty(
                    batch_dim, k, p, device=x.device, dtype=x.dtype
                )
                dx = (
                    torch.bmm(dout1, w1_bfly.conj(), out=dx.transpose(0, 1))
                    .transpose(0, 1)
                    .reshape(*batch_shape, n)
                )
            if ctx.needs_input_grad[1]:
                x_reshaped = x.reshape(batch_dim, k, p).transpose(0, 1)
                dw1_bfly = torch.bmm(dout1.transpose(-1, -2), x_reshaped.conj())
        return dx, dw1_bfly, dw2_bfly


blockdiag_butterfly_multiply = BlockdiagButterflyMultiply.apply


def blockdiag_weight_to_dense_weight(weight):
    """
    Argumments:
        weight: (nblocks, out / nblocks, in / blocks)
    Return:
        dense_weight: (out / in)
    """
    return torch.block_diag(*torch.unbind(weight, dim=0))


def blockdiag_multiply_reference(x, weight):
    """
    This implementation is slow but more likely to be correct.
    Arguments:
        x: (..., n)
        weight: (nblocks, q, n / nblocks)
    Outputs:
        out: (..., nblocks * q)
    """
    n = x.shape[-1]
    nblocks, q, p = weight.shape
    assert nblocks * p == n

    x_reshaped = rearrange(
        x, "... (nblocks p) -> ... nblocks p", nblocks=nblocks
    )
    return rearrange(
        torch.einsum("...kp, kqp -> ...kq", x_reshaped, weight),
        "... nblocks q -> ... (nblocks q)",
    )


class BlockdiagMultiply(torch.autograd.Function):
    """This is a faster implementation, with careful memory copies for the fastest
    bmm performance.
    The backward pass is also written manually with careful memory copies.
    Arguments:
        x: (..., n)
        weight: (nblocks, q, n / nblocks)
    Outputs:
        out: (..., nblocks * q)
    """

    @staticmethod
    @torch.cuda.amp.custom_fwd(cast_inputs=torch.bfloat16)
    def forward(ctx, x, weight):
        ctx.save_for_backward(x, weight)
        batch_shape, n = x.shape[:-1], x.shape[-1]
        batch_dim = np.prod(batch_shape)
        nblocks, q, p = weight.shape
        assert nblocks * p == n
        x_reshaped = x.reshape(batch_dim, nblocks, p).transpose(0, 1)
        out = torch.empty(
            batch_dim, nblocks, q, device=x.device, dtype=x.dtype
        ).transpose(0, 1)
        out = torch.bmm(
            x_reshaped, weight.transpose(-1, -2), out=out
        ).transpose(0, 1)
        return out.reshape(*batch_shape, nblocks * q)

    @staticmethod
    @torch.cuda.amp.custom_bwd
    def backward(ctx, dout):
        x, weight = ctx.saved_tensors
        batch_shape, n = x.shape[:-1], x.shape[-1]
        batch_dim = np.prod(batch_shape)
        nblocks, q, p = weight.shape
        assert nblocks * p == n
        dx, dweight = None, None
        dout_reshaped = dout.reshape(batch_dim, nblocks, q).transpose(0, 1)
        if ctx.needs_input_grad[0]:
            dx = torch.empty(
                batch_dim, nblocks, p, device=x.device, dtype=x.dtype
            )
            dx = (
                torch.bmm(dout_reshaped, weight.conj(), out=dx.transpose(0, 1))
                .transpose(0, 1)
                .reshape(*batch_shape, n)
            )
        if ctx.needs_input_grad[1]:
            x_reshaped = x.reshape(batch_dim, nblocks, p).transpose(0, 1)
            dweight = torch.bmm(
                dout_reshaped.transpose(-1, -2), x_reshaped.conj()
            )
        return dx, dweight


blockdiag_multiply = BlockdiagMultiply.apply


# Copyright (c) 2023, Dan Fu and Simran Arora.
# Adapted from https://github.com/HazyResearch/safari/blob/main/src/models/sequence/hyena.py


def fftconv_ref(
    u_variable,
    k,
    D_variable,
    dropout_mask,
    gelu=True,
    k_rev=None,
    flashfft=None,
):
    # u.shape:   B H L
    seqlen = u_variable.shape[-1]

    if flashfft is not None:
        y = flashfft(u_variable.to(dtype=torch.bfloat16).contiguous(), k)
    else:
        fft_size = 2 * seqlen
        k_f = torch.fft.rfft(k, n=fft_size) / fft_size
        if k_rev is not None:
            k_rev_f = torch.fft.rfft(k_rev, n=fft_size) / fft_size
            k_f = k_f + k_rev_f.conj()
        u_f = torch.fft.rfft(u_variable.to(dtype=k.dtype), n=fft_size)

        if len(u_variable.shape) > 3:
            k_f = k_f.unsqueeze(1)

        y = torch.fft.irfft(u_f * k_f, n=fft_size, norm="forward")[..., :seqlen]

    out = y + u_variable * D_variable

    if gelu:
        out = F.gelu(out)
    if dropout_mask is not None:
        return (out * rearrange(dropout_mask, "b H -> b H 1")).to(
            dtype=u_variable.dtype
        )
    else:
        return out.to(dtype=u_variable.dtype)


@torch.jit.script
def mul_sum(q, y):
    return (q * y).sum(dim=1)


class Sin(nn.Module):
    def __init__(self, dim, w=10, w_mod=1, train_freq=True):
        super().__init__()

        init_tensor = torch.ones(1, dim)
        self.freq = (
            nn.Parameter(w * init_tensor)
            if train_freq
            else w * torch.ones(1, dim)
        )
        self.w_mod = w_mod

    def forward(self, x):
        return torch.sin(self.w_mod * self.freq * x)


class StructuredLinear(nn.Module):
    def __init__(
        self, in_features, out_features, bias=True, device=None, dtype=None
    ):
        """Subclasses should call reset_parameters"""
        factory_kwargs = {"device": device, "dtype": dtype}
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        # Subclasses may override {in,out}_features_extended
        if not hasattr(self, "in_features_extended"):
            self.in_features_extended = in_features
        if not hasattr(self, "out_features_extended"):
            self.out_features_extended = out_features
        if bias:
            self.bias = nn.Parameter(
                torch.zeros(out_features, **factory_kwargs)
            )
        else:
            self.register_parameter("bias", None)

    def reset_parameters(self) -> None:
        self.set_weights_from_dense_init(
            dense_init_fn_=partial(init.kaiming_uniform_, a=math.sqrt(5))
        )
        self.reset_parameters_bias()

    def set_weights_from_dense_init(self, dense_init_fn_):
        raise NotImplementedError

    def reset_parameters_bias(self):
        if self.bias is not None:
            fan_in = self.bias.shape[-1]
            bound = 1 / math.sqrt(fan_in) if fan_in > 0 else 0
            init.uniform_(self.bias, -bound, bound)

    @property
    def saving(self):
        raise NotImplementedError

    def convert_to_dense_weight(self):
        factory_kwargs = {
            "device": self.weight.device,
            "dtype": self.weight.dtype,
        }
        dense_weight = self.forward_matmul(
            torch.eye(self.in_features, **factory_kwargs)
        ).T
        return dense_weight

    def preprocess(self, x):
        in_features = x.shape[-1]
        if in_features < self.in_features_extended:
            x = F.pad(x, (0, self.in_features_extended - in_features))
        return x

    def postprocess(self, output):
        out_features_extended = output.shape[-1]
        if out_features_extended > self.out_features:
            output = output[..., : self.out_features]
        return output

    def forward_matmul(self, x):
        raise NotImplementedError

    def forward(self, x):
        output = self.forward_matmul(x)
        # Convert bias to output.dtype in case of AMP, otherwise bias and activation will be in FP32
        return (
            (output + self.bias.to(dtype=output.dtype))
            if self.bias is not None
            else output
        )
